from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import torch
from torch import nn


@dataclass
class Permutation:
    """A permutation."""
    perm_indices: torch.Tensor

    def __len__(self) -> int:
        return len(self.perm_indices)


@dataclass
class ModuleParameters:
    weight_a: nn.Parameter | torch.Tensor
    weight_b: nn.Parameter | torch.Tensor
    name: str
    axis_to_permutation: dict[int, Permutation]
    module_type: Any

    # A parameter is either only a weight, or a weight with associated bias.
    # This is because the weights are used for associating permutations.
    # A bias is always associated with axis 0, the output axis.
    bias_a: nn.Parameter | None = None
    bias_b: nn.Parameter | None = None

    @property
    def input_permutation(self) -> Permutation:
        if len(self.axis_to_permutation) == 1:
            return self.axis_to_permutation[0]
        else:
            return self.axis_to_permutation[1]

    @input_permutation.setter
    def input_permutation(self, perm: Permutation) -> None:
        if len(self.axis_to_permutation) == 1:
            self.axis_to_permutation[0] = perm
        else:
            self.axis_to_permutation[1] = perm

    @property
    def output_permutation(self) -> Permutation:
        return self.axis_to_permutation[0]

    @output_permutation.setter
    def output_permutation(self, perm: Permutation) -> None:
        if self.module_type is nn.LayerNorm and len(self.axis_to_permutation) > 1:
            self.axis_to_permutation[0] = Permutation(torch.arange(len(perm)))
        else:
            self.axis_to_permutation[0] = perm

    def apply_permutations(self, except_axis: int = -1) -> None:
        for axis, permutation in self.axis_to_permutation.items():
            if axis == except_axis:
                continue
            self.permute_axis(self.weight_b, axis, permutation.perm_indices)
            if self.bias_b is not None and axis == 0:  # axis 0 is output dim -> bias
                self.permute_axis(self.bias_b, axis, permutation.perm_indices)

    @staticmethod
    @torch.no_grad()
    def permute_axis(
            param: nn.Parameter | torch.Tensor, axis: int, perm_indices: torch.Tensor
    ) -> None:
        """
        Permute a parameter along a given axis.

        Args:
            param:
                The parameter to permute.
            axis:
                The axis along which to permute.
            perm_indices:
                The permutation indices.
        """
        perm_indices = perm_indices.to(param.device)
        x = param.data.moveaxis(axis, 0)
        x = x[perm_indices]
        x = x.moveaxis(0, axis)
        param.data = x



@dataclass
class ModelInfo:
    """
    Information about a model.

    Attributes:
        name:
            The name of the model.
        contains_batch_norm:
            Whether the model contains BatchNorms.
            This is relevant because BatchNorms have to have their
            statistics recalculated after a permutation & interpolation.
        num_parameters:
            The total number of parameters in the model.
        num_weights:
            The number of weight parameters in the model.
        num_elements_per_weight:
            The number of elements in each weight parameter as a list.
        num_permutable_elements_per_weight:
            The number of elements in each weight parameter that can be permuted.
            This is the size of axes 0 and 1 for 2d-weights,
            and the size of axis 0 for 1d-weights.
        num_biases:
            The number of bias parameters in the model.
        num_elements_per_bias:
            The number of elements in each bias parameter. All are permutable.
        num_elements_per_weight_mean:
            The mean number of elements in each weight parameter.
        num_elements_per_bias_mean:
            The mean number of elements in each bias parameter.
        num_elements_per_weight_std:
            The standard deviation of the number of elements in each weight parameter.
        num_elements_per_bias_std:
            The standard deviation of the number of elements in each bias parameter.
    """
    name: str
    contains_batch_norm: bool
    num_parameters: int
    num_weights: int
    num_elements_per_weight: list[int]
    num_permutable_elements_per_weight: list[int]
    num_biases: int
    num_elements_per_bias: list[int]
    num_elements_per_weight_mean: float
    num_elements_per_bias_mean: float
    num_elements_per_weight_std: float
    num_elements_per_bias_std: float

    def __repr__(self) -> str:
        return (
            f"ModelInfo: "
            f"\n\tname: {self.name}"
            f"\n\tcontains_batch_norm: {self.contains_batch_norm}"
            f"\n\tnum_parameters: {self._pretty_num_str(self.num_parameters)}"
            f"\n\tnum_weights: {self._pretty_num_str(self.num_weights)}"
            f"\n\tnum_elements_per_weight: "
            f"{self._pretty_list_str(self.num_elements_per_weight)}"
            f"\n\tnum_permutable_elements_per_weight: "
            f"{self._pretty_list_str(self.num_permutable_elements_per_weight)}"
            f"\n\tnum_biases: {self._pretty_num_str(self.num_biases)}"
            f"\n\tnum_elements_per_bias: "
            f"{self._pretty_list_str(self.num_elements_per_bias)}"
            f"\n\tnum_elements_per_weight_mean: "
            f"{self._pretty_num_str(self.num_elements_per_weight_mean)}"
            f"\n\tnum_elements_per_bias_mean: "
            f"{self._pretty_num_str(self.num_elements_per_bias_mean)}"
            f"\n\tnum_elements_per_weight_std: "
            f"{self._pretty_num_str(self.num_elements_per_weight_std)}"
            f"\n\tnum_elements_per_bias_std: "
            f"{self._pretty_num_str(self.num_elements_per_bias_std)}"
        )

    def __str__(self) -> str:
        return (
            f"ModelInfo("
            f"name={self.name}, "
            f"contains_batch_norm={self.contains_batch_norm}, "
            f"num_parameters={self._pretty_num_str(self.num_parameters)}, "
            f"num_weights={self._pretty_num_str(self.num_weights)}, "
            f"num_elements_per_weight="
            f"{self._pretty_list_str(self.num_elements_per_weight)}, "
            f"num_permutable_elements_per_weight="
            f"{self._pretty_list_str(self.num_permutable_elements_per_weight)}, "
            f"num_biases={self._pretty_num_str(self.num_biases)}, "
            f"num_elements_per_bias="
            f"{self._pretty_list_str(self.num_elements_per_bias)}, "
            f"num_elements_per_weight_mean="
            f"{self._pretty_num_str(self.num_elements_per_weight_mean)}, "
            f"num_elements_per_bias_mean="
            f"{self._pretty_num_str(self.num_elements_per_bias_mean)}, "
            f"num_elements_per_weight_std="
            f"{self._pretty_num_str(self.num_elements_per_weight_std)}, "
            f"num_elements_per_bias_std="
            f"{self._pretty_num_str(self.num_elements_per_bias_std)}"
            f")"
        )

    @staticmethod
    def _pretty_list_str(lst: list[int]) -> str:
        plist = [ModelInfo._pretty_num_str(x) for x in lst]
        if len(plist) > 3:
            plist = plist[:3] + ["..."]
        return "[" + ", ".join([num for num in plist]) + "]"

    @staticmethod
    def _pretty_num_str(num: int | float) -> str:
        rest = 0.0
        if isinstance(num, float):
            rest = num % 1
            num = int(num)

        numstr = f"{num:_}"

        if rest:
            numstr += f".{str(rest).split('.')[1][:3]}"

        return numstr



"""
The purpose of this test is to find out how to properly permute the weights such that
the output of the model stays constant.

This is for different Module-types.
"""

from __future__ import annotations

import copy
import itertools

import pytest
import torch
from torch import nn

from tests.fixtures.utils import allclose, model_change_percent


def test_model_output_consistency_tensors() -> None:
    """
    Demonstrate that the output of a model is invariant to the
    permutation of the weight, iff the input is also permuted accordingly.
    """
    W = torch.randn(10)
    x = torch.randn(10)
    y_orig = x @ W

    perm = torch.randperm(10)
    W_new = W[perm]
    x_new = x[perm]
    y_new = x_new @ W_new

    assert allclose(y_orig, y_new)


class TestConsistencyLinear:

    @staticmethod
    def test_model_output_consistency_lin_only_input() -> None:
        """
        Demonstrate that, as long as the output dimension is
        untouched, permuting the input and the input-dimension of the weights
        yields the same output as the unpermuted input and weight.
        """
        model = nn.Linear(10, 10)
        x = torch.randn(10)
        y_orig = model(x)

        perm = torch.randperm(10)
        model.weight.data = model.weight.data[:, perm]
        y_xfail = model(x)
        assert not torch.allclose(y_orig, y_xfail)

        x_new = x[perm]
        y_xpass = model(x_new)
        assert torch.allclose(y_orig, y_xpass, atol=1e-7, rtol=1e-4)

    @staticmethod
    def test_permuted_output_produces_same_digits() -> None:
        model = nn.Linear(10, 10, bias=False)
        x = torch.randn(10)
        y_orig = model(x)

        perm_out = torch.randperm(10)
        model.weight.data = model.weight.data[perm_out]
        y_new = model(x)

        for digit in y_new:
            assert digit in y_orig

    @staticmethod
    def test_perm_output_and_next_input_with_same_permutation() -> None:
        for _ in range(100):
            lin1, lin2 = nn.Linear(8, 10, bias=False), nn.Linear(10, 12, bias=False)
            model = nn.Sequential(lin1, lin2)
            x = torch.randn(8)
            y_orig = model(x)

            perm = torch.randperm(10)

            lin1.weight.data = lin1.weight.data[perm]
            lin2.weight.data = lin2.weight.data[:, perm]

            y_new = model(x)
            # Floating point operations lead to enough error that we need to
            # relax the tolerance a bit.
            assert allclose(y_orig, y_new)

    @staticmethod
    def test_perm_out_same_perm_in_multi_layer() -> None:
        for _ in range(100):
            lin1 = nn.Linear(8, 10, bias=False)
            lin2 = nn.Linear(10, 12, bias=False)
            lin3 = nn.Linear(12, 14, bias=False)

            model = nn.Sequential(lin1, lin2, lin3)
            x = torch.randn(8)
            y_orig = model(x)

            perm1 = torch.randperm(10)
            perm2 = torch.randperm(12)

            lin1.weight.data = lin1.weight.data[perm1]  # permute output dim of lin1
            lin2.weight.data = lin2.weight.data[:, perm1]  # reverse effects of perm1
            lin2.weight.data = lin2.weight.data[perm2]  # permute output dim of lin2
            lin3.weight.data = lin3.weight.data[:, perm2]  # reverse effects of perm2

            y_new = model(x)

            assert allclose(y_orig, y_new)

    @staticmethod
    def test_two_layer_control() -> None:
        lin1 = nn.Linear(8, 10, bias=False)
        lin2 = nn.Linear(10, 8, bias=False)

        model = nn.Sequential(lin1, lin2)
        x = torch.randn(8)
        y_orig = model(x)

        perm1 = torch.tensor([1, 2, 5, 3, 8, 9, 0, 6, 7, 4])
        perm2 = torch.tensor([4, 7, 2, 6, 0, 3, 1, 9, 5, 8])

        lin1.weight.data = lin1.weight.data[perm1]
        lin2.weight.data = lin2.weight.data[:, perm2]

        # This shouldn't work if lin1-output_perm != lin2-input_perm
        y_new = model(x)
        assert not allclose(y_orig, y_new)


class TestConsistencyConv1d:

    @staticmethod
    def test_output_dim_perm_is_output_perm() -> None:
        conv = nn.Conv1d(3, 2, 3, bias=False)
        x = torch.randn(1, 3, 9)
        y_orig = conv(x)

        perm = torch.tensor([1, 0])
        conv.weight.data = conv.weight.data[perm]
        y_new = conv(x)

        assert torch.allclose(y_orig, y_new[:, perm])

    @staticmethod
    def test_output_perm_multi_layer() -> None:
        for _ in range(10):
            conv1 = nn.Conv1d(3, 10, 3, bias=False)
            conv2 = nn.Conv1d(10, 12, 3, bias=False)
            relu = nn.ReLU()

            model = nn.Sequential(conv1, relu, conv2)
            x = torch.randn(4, 3, 9)
            y_orig = model(x)

            perm = torch.randperm(10)
            conv1.weight.data = conv1.weight.data[perm]
            conv2.weight.data = conv2.weight.data[:, perm]

            y_new = model(x)

            assert allclose(y_orig, y_new)


class TestConsistencyConv2d:

    @staticmethod
    def test_output_dim_perm_is_output_perm() -> None:
        conv = nn.Conv2d(3, 2, (3, 3), bias=False)
        x = torch.randn(1, 3, 9, 9)
        y_orig = conv(x)

        perm = torch.tensor([1, 0])
        conv.weight.data = conv.weight.data[perm]
        y_new = conv(x)

        assert torch.allclose(y_orig, y_new[:, perm])

    @staticmethod
    def test_output_perm_multi_layer() -> None:
        for _ in range(10):
            # conv1.weight.shape == [10, 3, 3, 3]
            conv1 = nn.Conv2d(3, 10, (3, 3), bias=False)
            # conv2.weight.shape == [12, 10, 3, 3]
            conv2 = nn.Conv2d(10, 12, (3, 3), bias=False)
            relu = nn.ReLU()

            model = nn.Sequential(conv1, relu, conv2)
            x = torch.randn(4, 3, 9, 9)
            y_orig = model(x)

            perm = torch.randperm(10)
            conv1.weight.data = conv1.weight.data[perm]
            conv2.weight.data = conv2.weight.data[:, perm]

            y_new = model(x)

            assert allclose(y_orig, y_new)


class TestConsistencyConv3d:

    @staticmethod
    def test_output_dim_perm_is_output_perm() -> None:
        for in_channels, out_channels, kernel_size in itertools.product(
                [1, 3], [1, 2], [2, 3]
        ):
            conv = nn.Conv3d(in_channels, out_channels, kernel_size, bias=False)
            x = torch.randn(1, in_channels, 5, 5, 5)
            y_orig = conv(x)

            perm = torch.randperm(out_channels)
            conv.weight.data = conv.weight.data[perm]
            y_new = conv(x)

            assert torch.allclose(y_orig, y_new[:, perm])

    @staticmethod
    def test_output_perm_multi_layer() -> None:
        for in_channels, _out_channels, kernel_size in itertools.product(
                [1, 3], [1, 2], [2, 3]
        ):
            conv1 = nn.Conv3d(in_channels, 6, kernel_size, bias=False)
            conv2 = nn.Conv3d(6, 4, kernel_size, bias=False)
            relu = nn.ReLU()
            model = nn.Sequential(conv1, relu, conv2)
            x = torch.randn(4, in_channels, 5, 5, 5)
            y_orig = model(x)

            perm = torch.randperm(6)
            conv1.weight.data = conv1.weight.data[perm]
            conv2.weight.data = conv2.weight.data[:, perm]

            y_new = model(x)

            assert allclose(y_orig, y_new)


class TestLayerNorm:
    @staticmethod
    def test_multiplication_function() -> None:
        ln_norm = nn.LayerNorm(10)
        ln_weight = nn.LayerNorm(10)
        ln_weight.weight.data *= torch.randn(10)

        x = torch.randn(10)

        y_norm = ln_norm(x)
        y_weight = ln_weight(x)

        # LayerNorm normalizes, and then multiplies **element-wise**!
        assert torch.allclose(y_norm * ln_weight.weight.data, y_weight)

    @staticmethod
    def test_layernorm_after_linear() -> None:
        """If a model ends in a LayerNorm, then output has to be permuted
        by the inverse of the permutation that was applied to the weights
        of the LayerNorm."""
        lin = nn.Linear(10, 10, bias=False)
        ln = nn.LayerNorm(10)
        # Adjust LayerNorm weight to be non-identity
        ln.weight.data *= torch.randn(10)
        model = nn.Sequential(lin, ln)

        x = torch.randn(10)
        y_orig = model(x)

        perm = torch.randperm(10)
        lin.weight.data = lin.weight.data[perm]
        ln.weight.data = ln.weight.data[perm]
        y_new = model(x)
        rev_perm = torch.argsort(perm)
        y_new = y_new[rev_perm]

        assert allclose(y_orig, y_new)

    @staticmethod
    def test_layernorm_before_linear() -> None:
        ln = nn.LayerNorm(10)
        # Adjust LayerNorm weight to be non-identity
        ln.weight.data *= torch.randn(10)
        lin = nn.Linear(10, 10, bias=False)
        model = nn.Sequential(ln, lin)

        x = torch.randn(10)
        y_orig = model(x)

        perm = torch.randperm(10)
        ln.weight.data = ln.weight.data[perm]
        lin.weight.data = lin.weight.data[:, perm]

        y_new = model(x[perm])  # permute input to match permutation of weights

        assert allclose(y_orig, y_new)

    @staticmethod
    def test_layernorm_between_linears() -> None:
        lin1 = nn.Linear(10, 10, bias=False)
        ln = nn.LayerNorm(10)
        lin2 = nn.Linear(10, 10, bias=False)
        # Adjust LayerNorm weight to be non-identity
        ln.weight.data *= torch.randn(10)
        model = nn.Sequential(lin1, ln, lin2)

        x = torch.randn(10)
        y_orig = model(x)

        perm = torch.randperm(10)
        lin1.weight.data = lin1.weight.data[perm]
        ln.weight.data = ln.weight.data[perm]
        lin2.weight.data = lin2.weight.data[:, perm]
        y_new = model(x)

        assert allclose(y_orig, y_new)

    @staticmethod
    def test_layernorm_before_conv() -> None:
        for _ in range(10):
            x = torch.randn(2, 3, 18, 18)

            ln = nn.LayerNorm([2, 3, 18, 18])
            ln.weight.data *= torch.randn(2, 3, 18, 18)
            conv = nn.Conv2d(3, 4, (3, 3), bias=False)

            model = nn.Sequential(ln, conv)
            y_orig = model(x)

            perm_in = torch.randperm(3)
            ln.weight.data = ln.weight.data[:, perm_in]
            conv.weight.data = conv.weight.data[:, perm_in]
            y_new = model(x[:, perm_in])

            assert allclose(y_orig, y_new)

    @staticmethod
    def test_layernorm_after_conv() -> None:
        for _ in range(10):
            x = torch.randn(2, 3, 4, 4)

            conv = nn.Conv2d(3, 4, (3, 3), bias=False)
            ln = nn.LayerNorm([2, 4, 2, 2])
            ln.weight.data *= torch.randn(2, 4, 2, 2)

            model = nn.Sequential(conv, ln)
            y_orig = model(x)

            perm_out = torch.randperm(4)
            conv.weight.data = conv.weight.data[perm_out]
            ln.weight.data = ln.weight.data[:, perm_out]
            y_new = model(x)

            assert allclose(y_orig, y_new[:, torch.argsort(perm_out)])

    @staticmethod
    def test_layernorm_between_conv() -> None:
        for _ in range(10):
            x = torch.randn(2, 3, 6, 6)

            conv1 = nn.Conv2d(3, 4, (3, 3), bias=False)
            ln = nn.LayerNorm([2, 4, 4, 4])
            ln.weight.data *= torch.randn(2, 4, 4, 4)
            conv2 = nn.Conv2d(4, 5, (3, 3), bias=False)

            model = nn.Sequential(conv1, ln, conv2)
            y_orig = model(x)

            perm = torch.randperm(4)
            conv1.weight.data = conv1.weight.data[perm]
            ln.weight.data = ln.weight.data[:, perm]
            conv2.weight.data = conv2.weight.data[:, perm]
            y_new = model(x)

            assert allclose(y_orig, y_new)


class TestBatchNorm1d:
    @staticmethod
    def test_multiplication_function() -> None:
        bn_norm = nn.BatchNorm1d(10)
        bn_weight = nn.BatchNorm1d(10)
        bn_weight.weight.data *= torch.randn(10)

        x = torch.randn(3, 10)

        y_norm = bn_norm(x)
        y_weight = bn_weight(x)

        # BatchNorm normalizes, and then multiplies **element-wise**!
        assert torch.allclose(y_norm * bn_weight.weight.data, y_weight)

    @staticmethod
    def test_batchnorm_after_linear() -> None:
        """If a model ends in a BatchNorm1d, then output has to be permuted
        by the inverse of the permutation that was applied to the weights
        of the BatchNorm1d.

        However, in BatchNorm1d, that inverse has to be applied to the
        input-dimension of the output!
        """
        lin = nn.Linear(10, 10, bias=False)
        bn = nn.BatchNorm1d(10)
        # Adjust BatchNorm weight to be non-identity
        bn.weight.data *= torch.randn(10)
        model = nn.Sequential(lin, bn)

        x = torch.randn(3, 10)
        y_orig = model(x)

        perm = torch.randperm(10)
        lin.weight.data = lin.weight.data[perm]
        bn.weight.data = bn.weight.data[perm]
        bn.reset_running_stats()
        y_new = model(x)
        rev_perm = torch.argsort(perm)
        y_new = y_new[:, rev_perm]

        assert allclose(y_orig, y_new)

    @staticmethod
    def test_batchnorm_before_linear() -> None:
        bn = nn.BatchNorm1d(10)
        # Adjust BatchNorm weight to be non-identity
        bn.weight.data *= torch.randn(10)
        lin = nn.Linear(10, 10, bias=False)
        model = nn.Sequential(bn, lin)

        x = torch.randn(3, 10)
        y_orig = model(x)

        bn.reset_running_stats()
        perm = torch.randperm(10)
        bn.weight.data = bn.weight.data[perm]
        lin.weight.data = lin.weight.data[:, perm]

        y_new = model(x[:, perm])

        assert allclose(y_orig, y_new)

    @staticmethod
    def test_batchnorm_between_linears() -> None:
        lin1 = nn.Linear(10, 10, bias=False)
        bn = nn.BatchNorm1d(10)
        lin2 = nn.Linear(10, 10, bias=False)
        # Adjust BatchNorm weight to be non-identity
        bn.weight.data *= torch.randn(10)
        model = nn.Sequential(lin1, bn, lin2)

        x = torch.randn(3, 10)
        y_orig = model(x)

        perm = torch.randperm(10)
        lin1.weight.data = lin1.weight.data[perm]
        bn.weight.data = bn.weight.data[perm]
        lin2.weight.data = lin2.weight.data[:, perm]
        bn.reset_running_stats()
        y_new = model(x)

        assert allclose(y_orig, y_new)


class TestBatchNorm2d:
    @staticmethod
    def test_batchnorm_after_conv2d() -> None:
        conv = nn.Conv2d(3, 10, 3, bias=False)
        bn = nn.BatchNorm2d(10)
        model = nn.Sequential(conv, bn)

        x = torch.randn(4, 3, 9, 9)
        y_orig = model(x)

        perm = torch.randperm(10)
        conv.weight.data = conv.weight.data[perm]
        bn.weight.data = bn.weight.data[perm]
        bn.reset_running_stats()
        y_new = model(x)
        rev_perm = torch.argsort(perm)
        y_new = y_new[:, rev_perm]

        assert allclose(y_orig, y_new)

    @staticmethod
    def test_batchnorm_before_conv2d() -> None:
        bn = nn.BatchNorm2d(3)
        conv = nn.Conv2d(3, 10, 3, bias=False)
        model = nn.Sequential(bn, conv)

        x = torch.randn(4, 3, 9, 9)
        y_orig = model(x)

        bn.reset_running_stats()
        perm = torch.randperm(3)
        bn.weight.data = bn.weight.data[perm]
        conv.weight.data = conv.weight.data[:, perm]

        y_new = model(x[:, perm])

        assert allclose(y_orig, y_new)


class TestConsistencyBatchNorm3d:

    @staticmethod
    def test_multiplication_function() -> None:
        for num_channels in [1, 3, 5]:
            bn_norm = nn.BatchNorm3d(num_channels)
            bn_weight = nn.BatchNorm3d(num_channels)
            bn_weight.weight.data *= torch.randn(num_channels)

            x = torch.randn(4, num_channels, 5, 5, 5)

            y_norm = bn_norm(x)
            y_weight = bn_weight(x)

            # BatchNorm3d normalizes, and then multiplies **element-wise**!
            assert torch.allclose(
                y_norm * bn_weight.weight.data.view(1, num_channels, 1, 1, 1),
                y_weight
            )

    @staticmethod
    def test_batchnorm_after_conv3d() -> None:
        for in_channels, out_channels, kernel_size in itertools.product(
                [1, 3], [1, 2], [2, 3]
        ):
            conv = nn.Conv3d(in_channels, out_channels, kernel_size, bias=False)
            bn = nn.BatchNorm3d(out_channels)
            # Adjust BatchNorm weight to be non-identity
            bn.weight.data *= torch.randn(out_channels)
            model = nn.Sequential(conv, bn)

            x = torch.randn(4, in_channels, 5, 5, 5)
            y_orig = model(x)

            perm = torch.randperm(out_channels)
            conv.weight.data = conv.weight.data[perm]
            bn.weight.data = bn.weight.data[perm]
            bn.reset_running_stats()
            y_new = model(x)
            rev_perm = torch.argsort(perm)
            y_new = y_new[:, rev_perm]

            assert allclose(y_orig, y_new)

    @staticmethod
    def test_batchnorm_before_conv3d() -> None:
        for in_channels, out_channels, kernel_size in itertools.product(
                [1, 3], [1, 2], [2, 3]
        ):
            bn = nn.BatchNorm3d(in_channels)
            # Adjust BatchNorm weight to be non-identity
            bn.weight.data *= torch.randn(in_channels)
            conv = nn.Conv3d(in_channels, out_channels, kernel_size, bias=False)
            model = nn.Sequential(bn, conv)

            x = torch.randn(4, in_channels, 5, 5, 5)
            y_orig = model(x)

            bn.reset_running_stats()
            perm = torch.randperm(in_channels)
            bn.weight.data = bn.weight.data[perm]
            conv.weight.data = conv.weight.data[:, perm]

            y_new = model(x[:, perm])

            assert allclose(y_orig, y_new)

    @staticmethod
    def test_batchnorm_between_conv3d() -> None:
        for in_channels, mid_channels, out_channels, kernel_size in itertools.product(
                [1, 3], [4], [1, 2], [2, 3]
        ):
            conv1 = nn.Conv3d(in_channels, mid_channels, kernel_size, bias=False)
            bn = nn.BatchNorm3d(mid_channels)
            conv2 = nn.Conv3d(mid_channels, out_channels, kernel_size, bias=False)
            # Adjust BatchNorm weight to be non-identity
            bn.weight.data *= torch.randn_like(bn.weight.data)
            model = nn.Sequential(conv1, bn, conv2)

            x = torch.randn(4, in_channels, 5, 5, 5)
            y_orig = model(x)

            bn.reset_running_stats()
            perm = torch.randperm(mid_channels)
            bn.weight.data = bn.weight.data[perm]
            conv2.weight.data = conv2.weight.data[:, perm]
            conv1.weight.data = conv1.weight.data[perm]

            y_new = model(x)

            assert allclose(y_orig, y_new)


class TestMultiheadAttention:
    @staticmethod
    def test_io() -> None:
        embed_dim = 6
        num_heads = 3

        class Model(nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.mha = nn.MultiheadAttention(embed_dim, num_heads, bias=False)
                self.lin1 = nn.Linear(embed_dim, embed_dim, bias=False)
                self.lin2 = nn.Linear(embed_dim, embed_dim, bias=False)
                self.relu = nn.ReLU()

            def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:
                input_tensor = self.relu(self.lin1(input_tensor))
                input_tensor = self.relu(
                    self.mha(input_tensor, input_tensor, input_tensor)[0]
                )
                input_tensor = self.relu(self.lin2(input_tensor))
                return input_tensor

        model = Model()
        model_orig = copy.deepcopy(model)

        x = torch.randn(embed_dim, embed_dim)
        y_orig = model(x)

        perm1 = torch.tensor([0, 2, 4, 3, 5, 1], dtype=torch.long)
        perm2 = torch.tensor([2, 4, 5, 3, 0, 1], dtype=torch.long)

        model.lin1.weight.data = model.lin1.weight.data[perm1]
        model.mha.in_proj_weight.data = model.mha.in_proj_weight.data[:, perm1]
        model.mha.out_proj.weight.data = model.mha.out_proj.weight.data[perm2]
        model.lin2.weight.data = model.lin2.weight.data[:, perm2]

        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)

        assert allclose(y_orig, y_new)


class TestConsistencyPath:
    @staticmethod
    def test_path() -> None:
        size = 4
        relu = nn.ReLU()
        lin1 = nn.Linear(size, size, bias=False)

        lin_res1 = nn.Linear(size, size, bias=False)
        lin_res2 = nn.Linear(size, size, bias=False)

        residual_path = nn.Sequential(lin_res1, nn.ReLU(), lin_res2, nn.ReLU())

        lin2 = nn.Linear(size, size, bias=False)

        class ResBlock(nn.Module):
            def forward(self, x: torch.Tensor) -> torch.Tensor:
                return x + residual_path(x)  # type: ignore[no-any-return]

        model = nn.Sequential(lin1, relu, ResBlock(), lin2, relu)
        model_orig = copy.deepcopy(model)
        x = torch.randn(size)
        y_orig = model(x)

        perm1 = [1, 3, 0, 2]
        perm2 = [1, 2, 3, 0]

        lin1.weight.data = lin1.weight.data[perm1]
        lin_res1.weight.data = lin_res1.weight.data[:, perm1]
        lin_res1.weight.data = lin_res1.weight.data[perm2]
        lin_res2.weight.data = lin_res2.weight.data[:, perm2]
        lin_res2.weight.data = lin_res2.weight.data[perm1]
        lin2.weight.data = lin2.weight.data[:, perm1]

        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)

        assert allclose(y_orig, y_new)



from __future__ import annotations

from typing import Any

import torch
from torch import nn
from torch.utils.data import DataLoader, TensorDataset

from rebasin.utils import contains_parameter, get_inputs_labels, recalculate_batch_norms
from tests.fixtures.models import SaveCallCount


def test_recalculate_batch_norms() -> None:
    conv = nn.Conv2d(1, 1, 3)
    bn = nn.BatchNorm2d(1)
    scc = SaveCallCount()
    model = torch.nn.Sequential(conv, bn, scc)
    dataloader = DataLoader(TensorDataset(torch.rand(10, 1, 10, 10)), batch_size=10)

    # Test whether recalculate_batch_norms works
    assert isinstance(bn.running_mean, torch.Tensor)
    assert torch.all(bn.running_mean == torch.zeros_like(bn.running_mean))
    recalculate_batch_norms(model, dataloader, [0], None, False)
    assert torch.all(bn.running_mean != torch.zeros_like(bn.running_mean))
    assert scc.call_count > 0  # Test that scc works

    # Test whether recalculate_batch_norms works in eval mode and resets it afterward
    model.eval()
    bn.running_mean = torch.zeros_like(bn.running_mean)  # reset running_mean
    recalculate_batch_norms(model, dataloader, [0], None, False)
    assert model.training is False  # Reset to eval mode
    assert torch.all(bn.running_mean != torch.zeros_like(bn.running_mean))  # Did work

    # Test whether recalculate_batch_norms stops early
    #   if there are no BatchNorms in the model
    scc.call_count = 0
    model = torch.nn.Sequential(nn.Linear(3, 3), nn.Linear(3, 3), scc)
    recalculate_batch_norms(model, dataloader, 0, None, False)
    assert scc.call_count == 0


def test_get_inputs_labels() -> None:
    # Setup
    in1, in2, in3 = torch.rand(10, 3), torch.rand(10, 3), torch.rand(10, 3)
    out1, out2, out3 = torch.rand(10, 3), torch.rand(10, 3), torch.rand(10, 3)

    # For mypy:
    inputs: list[Any]
    labels: list[Any]
    batch: tuple[torch.Tensor, ...]

    # Test default indices
    batch = in1, out1
    inputs, labels = get_inputs_labels(batch)
    assert torch.allclose(inputs[0], in1)
    assert torch.allclose(labels[0], out1)

    # Test custom indices
    batch = in1, in2, out1, out2
    inputs, labels = get_inputs_labels(
        batch, input_indices=[0, 1], label_indices=[2, 3]
    )
    assert torch.allclose(inputs[0], in1)
    assert torch.allclose(inputs[1], in2)
    assert torch.allclose(labels[0], out1)
    assert torch.allclose(labels[1], out2)

    # Test mixed indices
    batch = in2, in3, in1, out2, out3, out1
    inputs, labels = get_inputs_labels(
        batch, input_indices=[2, 0, 1], label_indices=[5, 3, 4]
    )
    assert torch.allclose(inputs[0], in1)
    assert torch.allclose(inputs[1], in2)
    assert torch.allclose(inputs[2], in3)
    assert torch.allclose(labels[0], out1)
    assert torch.allclose(labels[1], out2)
    assert torch.allclose(labels[2], out3)


def test_contains_parameter() -> None:
    p1 = nn.Parameter(torch.rand(3, 3))
    p2 = nn.Parameter(torch.rand(3))
    p3 = nn.Parameter(torch.rand(5))

    assert contains_parameter((p1, p2), p1)
    assert contains_parameter((p1, p2), p2)
    assert not contains_parameter((p1, p2), p3)



from __future__ import annotations

import copy
from typing import Any

import torch
from scipy.optimize import linear_sum_assignment  # type: ignore[import]
from torch import nn
from tqdm import tqdm

from rebasin.modules import Permutation, PermutationInfo  # type: ignore[attr-defined]
from rebasin.permutation_initializer import PermutationInitializer


def calculate_progress(
        cost_mat: torch.Tensor,
        perm_old: torch.Tensor,
        perm_new: torch.Tensor,
        device: torch.device | str | None = None,
) -> bool:
    """
    Compute the progress of the permutation.

    Args:
        cost_mat:
            The cost matrix. Shape: (n, n).
        perm_old:
            The old permutation. Shape: (n,).
        perm_new:
            The new permutation. Shape: (n,).
        device:
            The device on which to perform the computation.

    Returns:
        :code:`True` if the new permutation moves the cost matrix
        closer to the identity matrix than the old permutation does,
        else :code:`False`.
    """
    cost_mat, perm_old, perm_new = (
        cost_mat.to(device), perm_old.to(device), perm_new.to(device)
    )

    # linear_sum_assignment finds the minimum cost by finding a permutation
    #   of the cost matrix that moves the highest values to the diagonal.
    # In other words, it tries to maximize the cost matrix's closeness
    #   to the identity matrix.
    # Therefore, we need to check whether the new permutation moves the
    #   cost matrix closer to the identity matrix compared to the old one.
    # This is done by taking the dot product of the permuted cost matrix
    #   with the identity matrix and then summing over the rows.
    # This is equivalent to picking the diagonal elements of the permuted
    #   cost matrix, and is achieved using torch.einsum.
    # In the end, the diagonal elements of the cost matrix are summed up.
    # The larger the sum, the closer the cost matrix is to the identity matrix,
    #   and the more progress has been made.
    eye_ = torch.eye(cost_mat.shape[0]).to(device)  # cost_mat is square
    optimality_old = torch.einsum("ij,ij->i", cost_mat[:, perm_old], eye_).sum()
    optimality_new = torch.einsum("ij,ij->i", cost_mat[:, perm_new], eye_).sum()

    progress = optimality_new.item() > optimality_old.item() + 1e-12
    return progress


class PermutationCoordinateDescent:
    r"""
    Implements the permutation coordinate descent algorithm.


    Args:
        model_a:
            The target model.

            Type: :class:`torch.nn.Module`.
        model_b:
            The model that will have its weights permuted.

            Type: :class:`torch.nn.Module`.
        input_data_b:
            The input data to the model_b.
            Used for tracing the model, so as long as the model can
            call its forward pass with this data, it can be anything.
            As this may be anything, this class cannot move it to any device.
            Therefore, it is the user's responsibility to ensure that
            :code:`input_data` is on the same device as :code:`model_b`.

            Type: Any.
        input_data_a:
            The input data to the model_a.
            The same comments apply as to :code:`input_data_b`.
            However, this argument is optional. If it is :code:`None`,
            then :code:`input_data_b` will be used for both models.
            Make sure that the device is correct in this case.

            Type: Any | None.
            Default: None.
        device_a:
            The device on which :code:`model_a` is located.
            Both :code:`model_a` and :code:`model_b` must
            either be :code:`None` or both be given.
            The purpose of this argument is to allow the user to
            use multiple GPUs if the models are so large that
            only one fits on a single GPU.

            Type: torch.device | str | None.
            Default: None.
        device_b:
            The device on which :code:`model_b` is located.

            Type: torch.device | str | None.
            Default: None.
        verbose:
            If True, progress will be printed to the console.

            Type: bool.
            Default: False.
    """

    def __init__(
            self,
            model_a: nn.Module,
            model_b: nn.Module,
            input_data_b: Any,
            input_data_a: Any | None = None,
            device_a: torch.device | str | None = None,
            device_b: torch.device | str | None = None,
            verbose: bool = False,
    ) -> None:
        self.model_b = model_b
        self.device_a = device_a
        self.device_b = device_b
        self.verbose = verbose

        if verbose:
            print("Initializing permutations...")

        self.pinit = PermutationInitializer(
            model_a, model_b, input_data_b, input_data_a
        )
        self.pinit.model_graph.enforce_identity()

        if verbose:
            print("Done.")

    def rebasin(self, max_iterations: int = 100) -> None:
        """
        Run the permutation coordinate descent algorithm.

        Args:
            max_iterations:
                The maximum number of iterations to run the algorithm for.
                Likely to converge much faster than this.

                Default: 100.
        """
        self.calculate_permutations(max_iterations)
        self.apply_permutations()

    def calculate_permutations(self, max_iterations: int = 100) -> None:
        """
        Run the permutation coordinate descent algorithm to calculate the permutations.

        Args:
            max_iterations:
                The maximum number of iterations.
        """
        # Calculate the permutations
        if self.verbose:
            print("Calculating permutations...")

        loop = tqdm(range(max_iterations), disable=not self.verbose)
        for iteration in loop:
            progress = self._calculate_permutations_step(loop)
            if not progress:
                if self.verbose:
                    print(f"Stopping early after {iteration + 1} steps.")
                break

    def _calculate_permutations_step(self, loop: tqdm[int]) -> bool:
        """
        Run one iteration of the permutation coordinate descent algorithm.

        Returns:
            True if the algorithm made progress, False otherwise.
        """
        progress = False

        for i in torch.randperm(len(self.pinit.model_graph.permutation_to_info)):
            # For mypy: give typehint
            perm_and_info: tuple[
                Permutation, list[PermutationInfo]
            ] = self.pinit.model_graph.permutation_to_info[i]

            perm, perm_infos = perm_and_info
            n = len(perm.perm_indices)
            cost_tensor = torch.zeros((n, n)).to(self.device_b)

            if self.verbose:
                loop.write("Calculating cost tensor...")
            for perm_info in perm_infos:
                # Copy so that the original parameter isn't moved to device_b.
                # If this were not done, then in one of the later steps,
                #   model_a would be almost completely on device_b.
                # This is a problem because we want the models to be able to live
                #   on different devices to allow for larger models
                #   by splitting their memory needs.
                # Also, we will apply some permutations, but don't want this
                #   to affect the original parameter, because we might do this
                #   again in another iteration.
                info = copy.deepcopy(perm_info)

                # Apply all permutations except the one we are currently working on.
                info.module.apply_permutations(except_axis=info.axis)
                w_a = info.parameter_a.to(self.device_b)
                w_b = info.parameter_b.to(self.device_b)

                # We want a square matrix as a cost tensor.
                # It should have shape (n, n).
                # To achieve this, we first move the axis of interest to the front.
                w_a = w_a.moveaxis(info.axis, 0) if len(w_a.shape) > 1 else w_a
                w_b = w_b.moveaxis(info.axis, 0) if len(w_b.shape) > 1 else w_b

                # Then, we reshape the tensor to (n, -1).
                # This means that all dimensions except the first one are flattened.
                # This way, multiplying w_a with w_b.T
                #   will result in a matrix of shape (n, n).
                # That matrix --- the cost_tensor --- encodes the similarity between
                #   w_a and w_b along the axis of interest.
                w_a = w_a.reshape(n, -1) if len(w_a.shape) > 1 else w_a
                w_b = w_b.reshape(n, -1).mT if len(w_b.shape) > 1 else w_b

                # We calculate the cost tensor by multiplying w_a with w_b.mT
                #   and adding it to the previous cost tensor.
                # We add the cost tensors because one permutation can apply to several
                #   weights and biases.
                # Adding the costs introduces a dependency between the permutations.
                # That dependency is desirable because it is important to not just
                #   bring the models close to each other layer by layer,
                #   but ideally to make the entire path through the model
                #   as similar as possible between the two.
                cost_tensor += w_a @ w_b

            if self.verbose:
                loop.write(f"Done. Cost matrix shape: {cost_tensor.shape}.")
                loop.write("Linear Sum Assignment...")
            # Calculate the ideal permutations of the rows and columns
            #   of the cost tensor in order to maximize its closeness to
            #   the identity matrix (see `calculate_progress`).
            # ri, ci: row-indices, column-indices
            ri, ci = linear_sum_assignment(
                cost_tensor.cpu().detach().numpy(), maximize=True
            )
            ri, ci = torch.from_numpy(ri), torch.from_numpy(ci)

            # Since we want to match features, we want to permute the columns only.
            # It should be the case that the rows naturally don't change.
            # If they do, that's a problem.
            assert torch.allclose(ri, torch.arange(n)), \
                "The rows of the cost tensor should not change."

            if self.verbose:
                loop.write("Done. \nProgress calculation...")
            # It is important to calculate the progress because
            #   the `_calculate_permutations_step`-method will be run several times.
            # If there were no interdependence between the permutations,
            #   this method would converge in one iteration.
            # However, there is interdependence,
            #   so we might need to run it several times.
            # If we calculate the progress and see that it is False,
            #   we can stop early.
            progress = progress or calculate_progress(
                cost_tensor,
                perm_old=perm.perm_indices,
                perm_new=ci,
                device=self.device_b
            )

            if self.verbose:
                loop.write(f"Done. Progress: {progress}\n")

            # Update the permutation.
            perm.perm_indices = ci

        return progress

    def apply_permutations(self) -> None:
        """
        Apply the calculated permutations to the model.
        """
        # Apply the permutations
        if self.verbose:
            print("Applying permutations...")

        self.pinit.model_graph.apply_permutations()



from __future__ import annotations

import copy
import sys

import pytest
import torch
from torchvision.models import (  # type: ignore[import]
    ViT_B_16_Weights,
    resnet18,
    vit_b_16,
)

from rebasin.permutation_initializer import PermutationInitializer
from tests.fixtures.models import MLP
from tests.fixtures.utils import (
    allclose,
    model_change_percent,
    reset_bn_running_stats,
    tensor_diff_perc,
)


def test_mlp() -> None:
    model_a, model_b = MLP(15, 10), MLP(15, 10)
    model_b_orig = copy.deepcopy(model_b)
    input_data = torch.randn(1, 15)
    y_orig = model_b(input_data)

    pinit = PermutationInitializer(model_a, model_b, input_data)
    assert len(pinit.model_graph) == 1

    for permutation, _ in pinit.model_graph.permutation_to_info:
        permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

    pinit.model_graph.enforce_identity()
    pinit.model_graph.apply_permutations()
    assert model_change_percent(model_b, model_b_orig) > 0.1

    y_new = model_b(input_data)
    assert allclose(y_orig, y_new)


def test_mlp_control() -> None:
    model_a, model_b_orig = MLP(15, 10), MLP(15, 10)
    model_b1 = copy.deepcopy(model_b_orig)
    model_b2 = copy.deepcopy(model_b_orig)

    input_data = torch.randn(1, 15)

    pinit1 = PermutationInitializer(model_a, model_b1, input_data)
    pinit2 = PermutationInitializer(model_a, model_b2, input_data)

    for permutation, _ in pinit1.model_graph.permutation_to_info:
        permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

    pinit1.model_graph.enforce_identity()
    pinit1.model_graph.apply_permutations()

    for permutation, _ in pinit2.model_graph.permutation_to_info:
        permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

    pinit2.model_graph.input_permutation = None
    pinit2.model_graph.output_permutation = None
    pinit2.model_graph.apply_permutations()

    diff1 = tensor_diff_perc(model_b1(input_data), model_b_orig(input_data))
    diff2 = tensor_diff_perc(model_b2(input_data), model_b_orig(input_data))

    assert diff1 < diff2


def test_resnet18() -> None:
    model_a, model_b = resnet18(), resnet18()
    model_b_orig = copy.deepcopy(model_b)
    input_data = torch.randn(1, 3, 224, 224)

    pinit = PermutationInitializer(model_a, model_b, input_data)

    for permutation, _ in pinit.model_graph.permutation_to_info:
        del _
        permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

    pinit.model_graph.enforce_identity()
    pinit.model_graph.apply_permutations()
    assert model_change_percent(model_b, model_b_orig) > 0.1

    assert (
            pinit.model_graph[1].input_permutation
            is pinit.model_graph[1].output_permutation
            is pinit.model_graph[0].output_permutation
    )
    assert (
            pinit.model_graph[3].input_permutation
            is pinit.model_graph[3][0].input_permutation
            is pinit.model_graph[3][1].input_permutation
    )

    for module in model_b.modules():
        if isinstance(module, torch.nn.BatchNorm2d):
            assert torch.allclose(
                module.running_mean,  # type: ignore[arg-type]
                torch.zeros_like(module.running_mean)  # type: ignore[arg-type]
            )
            assert torch.allclose(
                module.running_var,  # type: ignore[arg-type]
                torch.ones_like(module.running_var)  # type: ignore[arg-type]
            )

    # nn.BatchNorm really plays a number on the enforce_identity() function,
    #   so for resnet18, I have to pick a much more relaxed metric
    #   than for other models.
    assert tensor_diff_perc(model_b(input_data), model_b_orig(input_data)) < 0.15
    assert tensor_diff_perc(model_b(input_data), model_a(input_data)) > 1.0


@pytest.mark.skipif("--full-suite" not in sys.argv, reason="Compute intense")
def test_vit_b_16() -> None:
    """Test how it works with modules that have no :class:`nn.BatchNorm2d`."""
    model_a = vit_b_16(weights=ViT_B_16_Weights.IMAGENET1K_V1)
    model_b = vit_b_16(weights=ViT_B_16_Weights.IMAGENET1K_SWAG_LINEAR_V1)
    model_b_orig = copy.deepcopy(model_b)
    input_data = torch.randn(1, 3, 224, 224)
    y_orig = model_b(input_data)

    pinit = PermutationInitializer(model_a, model_b, input_data)

    for permutation, _ in pinit.model_graph.permutation_to_info:
        del _
        permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

    pinit.model_graph.enforce_identity()
    pinit.model_graph.apply_permutations()
    assert model_change_percent(model_b, model_b_orig) > 0.1

    y_new = model_b(input_data)

    assert tensor_diff_perc(y_new, y_orig) < 1e-5
    assert allclose(y_new, y_orig)
    assert not allclose(model_a(input_data), model_b(input_data))





from __future__ import annotations

import copy
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from rebasin.utils import recalculate_batch_norms


class Interpolation:
    """
    Interpolate between two models.

    The base class for all interpolation classes.
    This class is not meant to be used directly.

    Args:
        models:
            The models to interpolate between.
            This can be multiple models (see explanation above).

        eval_fn:
            The function to evaluate the models with.
            Usually, this simply calculates the loss for every batch in a
            validation or test dataloader, but it can be anything,
            including accuracy, or perplexity, etc.
            By using a :code:`eval_fn`, the user has maximum control.

            Its signature must be as follows:

            .. code-block:: python
                dev eval_fn(
                    model: nn.Module, device: torch.device | str | None
                ) -> float:
                    ...

            The :code:`model` argument is necessary
            because different models will have to be
            evaluated. The :code:`device` argument is necessary because the different
            models can be evaluated on different devices
            (see the :code:`devices` argument).

        eval_mode:
            The mode to use for evaluation. If "min", the model with the lowest
            metric is chosen, if "max", the model with the highest metric is chosen.

        train_dataloader:
            The dataloader to use for training the models.
            If this is given, then the :class:`Interpolation` subclass will check if
            the model has :code:`BatchNorm` layers and if so, will
            recalculate their :code:`running_mean` and :code:`running_var`
            statistics using the training data.

            This is unfortunately strongly recommended if your model includes
            these statistics.

        devices:
            Models may need a large amount of GPU-memory.
            To avoid running out of memory,
            the argument :code:`devices` can be used to specify which device
            each of the given models in :code:`models` is on.
            As each model is evaluated using the function given in :code:`eval_fn`,
            the corresponding device is passed to :code:`eval_fn` as well.

        device_interp:
            The device to use for the interpolation.
            If this is :code:`None`, no parameter will be moved
            to a different device for interpolation, and the
            interpolated model will be created on CPU.
            Again, this argument is useful for saving on memory.

        input_indices:
            If a training dataloader is given to :code:`train_dataloader`,
            and the model has :code:`BatchNorm` layers, then the
            :code:`running_mean` and :code:`running_var` statistics
            are recalculated using the training data.
            However, :class:`Interpolation` does not know which output from
            the dataloader is the input to the model. Sometimes,
            a model takes several inputs, sometimes, and input and a target
            are given by the dataloader, etc.
            To solve this, the argument :code:`input_indices` can be used
            to provide the indices of the inputs to the model.

            Can be an integer or a sequence of integers.

            The outputs of the training dataloader corresponding to the
            indices given here will be used as inputs to the models to
            recalculate the :code:`running_mean` and :code:`running_var`

        savedir:
            The directory to save the models in.
            If :code:`None`, the models are not saved.

        save_all:
            If :code:`True`, all models are saved,
            if :code:`False`, only the best model is saved.

        verbose:
            If :code:`True`, prints the progress of the interpolation.
    """
    def __init__(
            self,
            models: Sequence[torch.nn.Module],
            eval_fn: Any,
            eval_mode: str = "min",
            train_dataloader: DataLoader[Any] | None = None,
            devices: Sequence[torch.device | str] | None = None,
            device_interp: torch.device | str | None = None,
            input_indices: Sequence[int] | int = 0,
            savedir: Path | str | None = None,
            save_all: bool = False,
            verbose: bool = False,
    ) -> None:
        self._sanity_checks(
            models,
            eval_fn,
            eval_mode,
            train_dataloader,
            devices,
            device_interp,
            input_indices,
            savedir,
            save_all,
            verbose
        )
        if verbose:
            print("Setting up interpolation...")

        if savedir is not None and isinstance(savedir, str):
            savedir = Path(savedir)

        self.models = models
        self.eval_fn = eval_fn
        self.eval_mode = eval_mode
        self.idx_fn = torch.argmax if eval_mode == "max" else torch.argmin
        self.train_dataloader = train_dataloader
        self.devices: Sequence[torch.device | str | None] = (
            devices if devices is not None
            else [None] * len(models)  # type: ignore[list-item]
        )
        self.device_interp = device_interp
        self.input_indices = input_indices
        self.savedir = savedir
        self.save_all = save_all
        self.verbose = verbose

        if self.verbose:
            print("Evaluating given models...")

        self.metrics_models = [
            self.eval_fn(m, d)
            for m, d in tqdm(
                zip(self.models, self.devices), disable=not self.verbose
            )
        ]
        if self.verbose:
            print("Done.")
        self.metrics_interpolated: list[float] = []

        best_idx = int(self.idx_fn(torch.tensor(self.metrics_models)))
        self.best_metric = self.metrics_models[best_idx]
        self.best_model = self.models[best_idx]
        self.best_model_name = f"model_{best_idx}.pt"

    def interpolate(
            self,
            steps: int = 20,
            savedir: str | Path | None = None,
            save_all: bool = False
    ) -> None:
        """Interpolate between the models and save the best one or all."""

    @staticmethod
    def _sanity_checks(
            models: Sequence[torch.nn.Module],
            eval_fn: Any,
            eval_mode: str,
            train_dataloader: DataLoader[Any] | None,
            devices: Sequence[torch.device | str] | None,
            device_interp: torch.device | str | None,
            input_indices: Sequence[int] | int,
            savedir: Path | str | None,
            save_all: bool = False,
            verbose: bool = False,
    ) -> None:
        assert isinstance(models, Sequence), "Models must be a sequence"
        assert all(isinstance(model, nn.Module) for model in models), \
            "All models must be a subclass of nn.Module"

        assert callable(eval_fn), "Eval function must be callable"
        assert isinstance(eval_mode, str)
        assert eval_mode in ["min", "max"], "Eval mode must be 'min' or 'max'"

        assert isinstance(train_dataloader, DataLoader | type(None))

        if devices is not None:
            assert isinstance(devices, Sequence)
            assert all(
                isinstance(device, str | torch.device) for device in devices
            )
            assert len(devices) == len(models)

        assert isinstance(device_interp, str | torch.device | type(None))

        assert isinstance(input_indices, int | Sequence)
        if isinstance(input_indices, Sequence):
            assert all(isinstance(i, int) for i in input_indices)

        assert isinstance(savedir, Path | str | type(None))
        assert isinstance(save_all, bool)

        assert len(models) > 1, "Need at least two models to interpolate between"

        if save_all:
            assert savedir is not None

        assert isinstance(verbose, bool)


class LerpSimple(Interpolation):
    r"""
    Linear interpolation between two models.

    For two models and `s` steps, the interpolation looks like this
    (using :math:`m_a` to denote model number :math:`a`,
    and :math:`i_{ab,c}` to denote interpolation step `c`
    between models number `a` and `b`):

    |

    .. math::

        m_1 - i_{12,1} - ... - i_{12,s} - m_2

    |

    For `n` models, the same is repeated between every other pair of models,
    meaning:

    |

    .. math::
        m_1 - i_{12,1} - ... - i_{12,s} - m_2
        - i_{23,1} - ... - i_{23,s} - m_3 - ... - m_n

    |

    This means that the total number of interpolations is :math:`s \cdot (n - 1)`.

    Args:
        models:
            The models to interpolate between.
            This can be multiple models (see explanation above).

        eval_fn:
            The function to evaluate the models with.
            Usually, this simply calculates the loss for every batch in a
            validation or test dataloader, but it can be anything,
            including accuracy, or perplexity, etc.
            By using a :code:`eval_fn`, the user has maximum control.

            Its signature must be as follows:

            .. code-block:: python
                dev eval_fn(
                    model: nn.Module, device: torch.device | str | None
                ) -> float:
                    ...

            The :code:`model` argument is necessary
            because different models will have to be
            evaluated. The :code:`device` argument is necessary because the different
            models can be evaluated on different devices
            (see the :code:`devices` argument).

        eval_mode:
            The mode to use for evaluation. If "min", the model with the lowest
            metric is chosen, if "max", the model with the highest metric is chosen.

        train_dataloader:
            The dataloader to use for training the models.
            If this is given, then :class:`LerpSimple` will check if
            the model has :code:`BatchNorm` layers and if so, will
            recalculate their :code:`running_mean` and :code:`running_var`
            statistics using the training data.

            This is unfortunately strongly recommended if your model includes
            these statistics.

        devices:
            Models may need a large amount of GPU-memory.
            To avoid running out of memory,
            the argument :code:`devices` can be used to specify which device
            each of the given models in :code:`models` is on.
            As each model is evaluated using the function given in :code:`eval_fn`,
            the corresponding device is passed to :code:`eval_fn` as well.

        device_interp:
            The device to use for the interpolation.
            If this is :code:`None`, no parameter will be moved
            to a different device for interpolation, and the
            interpolated model will be created on CPU.
            Again, this argument is useful for saving on memory.

        input_indices:
            If a training dataloader is given to :code:`train_dataloader`,
            and the model has :code:`BatchNorm` layers, then the
            :code:`running_mean` and :code:`running_var` statistics
            are recalculated using the training data.
            However, :class:`LerpSimple` does not know which output from
            the dataloader is the input to the model. Sometimes,
            a model takes several inputs, sometimes, and input and a target
            are given by the dataloader, etc.
            To solve this, the argument :code:`input_indices` can be used
            to provide the indices of the inputs to the model.

            Can be an integer or a sequence of integers.

            The outputs of the training dataloader corresponding to the
            indices given here will be used as inputs to the models to
            recalculate the :code:`running_mean` and :code:`running_var`

        savedir:
            The directory to save the models in.
            If :code:`None`, the models are not saved.

        save_all:
            If :code:`True`, all models are saved,
            if :code:`False`, only the best model is saved.

        verbose:
            If :code:`True`, prints the progress of the interpolation.
    """

    def interpolate(
            self,
            steps: int = 20,
            savedir: str | Path | None = None,
            save_all: bool | None = None
    ) -> None:
        r"""
        Interpolate between the models and save the best one or all
        in :code:`self.best_model`.

        Args:
            steps:
                The number of steps to take between each pair of models.

            savedir:
                The directory to save the models in. Overwrites the :code:`savedir`
                argument passed to the constructor if not None.
                If both are :code:`None`, the models are not saved.

                If :code:`not None`, the following naming schema is used for the files:

                - :code:`"model_{model_num}.pt"`
                    if the model is one of the original models.
                    :code:`model_num` is the index of the model
                    in the :code:`models` argument.
                    The original models are not saved, except the best model.

                - :code:`"interp_models_{m_i}_{m_i+1}_perc_{percentage}.pt"`
                  if the model is interpolated.
                  :math:`m_i` and :math:`m_{i+1}` are the indices of the models
                  between which the interpolation is done.
                  `percentage` is the percentage of the way through the interpolation.
                  If it is small, the model is closer to :math:`m_i`,
                  and if it is large, the model is closer to :math:`m_{i+1}`.

            save_all:
                Whether to save all interpolated models, or just the best one.
                The best model is saved either way if a `savedir` is given.
                Overwrites the `save_all` argument passed to the constructor
                if not None.
        """
        # SANITY CHECKS AND DEFAULT SETTINGS
        assert isinstance(savedir, Path | str | type(None))
        if savedir is None:
            savedir = self.savedir
        elif isinstance(savedir, str):
            savedir = Path(savedir)

        if save_all is None:
            save_all = self.save_all

        assert isinstance(save_all, bool)

        if self.verbose:
            print("Interpolating between models...")

        # INTERPOLATION
        for step in tqdm(range(steps), disable=not self.verbose):
            # Interpolate between the two models
            # (step + 1) so that it never starts at zero percent.
            # (steps + 2) so that it never ends at 100 percent.
            # This is because the start and end models already exist
            percentage = (step + 1) / (steps + 2)
            self._interpolate_step(
                percentage=percentage, savedir=savedir, save_all=save_all
            )

        # SAVE THE BEST MODEL
        # If all interpolated models are saved and the best model is interpolated,
        #   it is unnecessary to save it again.
        # Conversely, if not all interpolated models are saved, the best model
        #   must be saved.
        # Alternatively, save_all only leads to interpolated models being saved,
        #   so the best model must be saved again if it is not interpolated.
        save_best_model = (not save_all) or (self.best_model in self.models)

        if save_best_model and (savedir is not None):
            torch.save(
                self.best_model.state_dict(),
                savedir.joinpath(self.best_model_name)
            )

    def _interpolate_step(
            self, percentage: float, savedir: Path | None, save_all: bool
    ) -> None:
        # Interpolate between all models
        for model_num, (model1, model2) in enumerate(
                zip(self.models[:-1], self.models[1:])
        ):
            model_interp = copy.deepcopy(model1).to(self.device_interp)

            for param1, param2, param_interp in zip(
                    model1.parameters(),
                    model2.parameters(),
                    model_interp.parameters(),
            ):
                param_interp.data = torch.lerp(
                    param1.to(self.device_interp),
                    param2.to(self.device_interp),
                    percentage
                )

            # Recalculate BatchNorm statistics
            if self.train_dataloader is not None:
                recalculate_batch_norms(
                    model_interp,
                    self.train_dataloader,
                    self.input_indices,
                    device=self.device_interp,
                    verbose=self.verbose
                )

            # Evaluate
            metric = self.eval_fn(model_interp, self.device_interp)
            self.metrics_interpolated.append(metric)

            # Save
            filename = f"interp_models_{model_num}_{model_num+1}" \
                       f"_perc_{percentage:.3f}.pt"

            is_best = (
                metric < self.best_metric
                if self.eval_mode == "min"
                else metric > self.best_metric
            )
            if is_best:
                self.best_metric = metric
                self.best_model = model_interp
                self.best_model_name = filename  # for later saving if not save_all

            if savedir is not None and save_all:
                torch.save(model_interp.state_dict(), savedir.joinpath(filename))



from __future__ import annotations

from collections.abc import Iterator

from rebasin.modules import (  # type: ignore[attr-defined]
    MODULE_TYPES,
    LayerNormModule,
    OneDimModule,
    Permutation,
    PermutationInfo,
)


class LinearPath:
    """A non-branching path through a model."""

    def __init__(self, *modules: MODULE_TYPES) -> None:
        self.modules = list(modules)
        for m in self.modules:
            assert isinstance(m, MODULE_TYPES), \
                "Linear paths must consist of modules from rebasin.modules."

    def __iter__(self) -> Iterator[MODULE_TYPES]:
        return iter(self.modules)

    def __len__(self) -> int:
        return len(self.modules)

    def __getitem__(self, index: int) -> MODULE_TYPES:
        return self.modules[index]

    def __bool__(self) -> bool:
        return bool(self.modules)

    @property
    def input_permutation(self) -> Permutation | None:
        """The permutation of the input to the first module."""
        return self[0].input_permutation if bool(self) else None

    @input_permutation.setter
    def input_permutation(self, permutation: Permutation | None) -> None:
        if not bool(self):
            return
        perm_pt = 0
        self[perm_pt].input_permutation = permutation

        # nn.Modules with 1D-weights have the same input- and output-permutation,
        # meaning that they have to be connected.
        while (
            perm_pt < len(self) - 1
            and isinstance(self[perm_pt], OneDimModule | LayerNormModule)
        ):
            self[perm_pt + 1].input_permutation = permutation
            perm_pt += 1

    @property
    def input_permutation_shape(self) -> int:
        return int(self[0].input_permutation_shape) if bool(self) else 0

    @property
    def input_shape(self) -> list[tuple[int, ...]]:
        return self[0].input_shape if bool(self) else []  # type: ignore[no-any-return]

    @property
    def output_permutation(self) -> Permutation | None:
        """The permutation of the output of the last module."""
        return self[-1].output_permutation if bool(self) else None

    @output_permutation.setter
    def output_permutation(self, permutation: Permutation | None) -> None:
        if not bool(self):
            return

        perm_pt = -1
        self[perm_pt].output_permutation = permutation

        while(
            perm_pt > -len(self)
            and isinstance(self[perm_pt], OneDimModule | LayerNormModule)
        ):
            self[perm_pt].input_permutation = permutation
            self[perm_pt - 1].output_permutation = permutation
            perm_pt -= 1

    @property
    def output_permutation_shape(self) -> int:
        return int(self[-1].output_permutation_shape) if bool(self) else 0

    @property
    def output_shape(self) -> list[tuple[int, ...]]:
        return (  # type: ignore[no-any-return]
            self[-1].output_shape
            if bool(self) else []
        )

    @property
    def permutation_to_info(self) -> list[tuple[Permutation, list[PermutationInfo]]]:
        id_to_perminfo: dict[int, tuple[Permutation, list[PermutationInfo]]] = {}

        for mod in self:
            for permutation, info in mod.permutation_to_info:
                if permutation is None:
                    continue
                if id(permutation) not in id_to_perminfo:
                    id_to_perminfo[id(permutation)] = (permutation, info)
                id_to_perminfo[id(permutation)][1].extend(info)

        return [perm_info for perm_info in id_to_perminfo.values()]

    def enforce_identity(
            self,
            prev_path: LinearPath | ParallelPaths | PathSequence | None = None,
            next_path: LinearPath | ParallelPaths | PathSequence | None = None
    ) -> None:
        """
        Enforce the constraint that the permutations in :class:`LinearPath`
        combined represent the identity function.

        In other words, applying the permutations in :class:`LinearPath`
        changes the layout of the weights and biases of the model, but not
        its output.
        """
        if not bool(self):
            return

        if prev_path is None:
            self.input_permutation = None
        elif prev_path.output_shape == self.input_shape:
            self.input_permutation = prev_path.output_permutation
        else:
            self.input_permutation = None
            prev_path.output_permutation = None

        pt0, pt1 = 0, 1
        while pt1 < len(self):
            if self[pt0].output_permutation is None:
                pt0 += 1
                if pt0 == pt1:
                    pt1 += 1
                continue
            elif self[pt1].input_permutation is None:
                pt1 += 1
                continue

            self[pt1].input_permutation = self[pt0].output_permutation
            pt0 += 1
            pt1 += 1

        if next_path is None:
            self.output_permutation = None

    def apply_permutations(self) -> None:
        """Apply the permutations in the path to the model."""
        for mod in self:
            mod.apply_permutations()

    def __repr__(self) -> str:
        modules_strings = [
            f"  {mod.__class__.__name__}("
            + f"\n    module.type: {mod.module_type.__name__}"
            + f"\n    input.shape: {mod.input_shape}"
            + f"\n    output.shape: {mod.output_shape}"
            + (
                f"\n    weight.in_dim.permutation.shape: {mod.input_permutation_shape}"
                if mod.input_permutation is not None
                else "\n    weight.in_dim.permutation: None"
            )
            + (
                f"\n    weight.out_dim.permutation.shape: "
                f"{mod.output_permutation_shape}"
                if mod.output_permutation is not None
                else "\n    weight.out_dim.permutation: None"
            )
            + f"\n  )\n\n"
            for mod in self
        ]

        if not modules_strings:
            return "\nLinearPath()\n"

        modules_strings = ["LinearPath(", *modules_strings] + [")"]

        # Normalize line width
        max_line_width = (
            max(len(line) for lines in modules_strings for line in lines.splitlines())
        )
        for i, lines in enumerate(modules_strings):
            normalized_lines = [
                line + " " * (max_line_width - len(line))
                for line in lines.splitlines()
            ]
            modules_strings[i] = "\n" + "\n".join(normalized_lines)

        return "".join(modules_strings)


class ParallelPaths:
    """
    Parallel paths.

    Paths through a model that all start and end in the same node
    are parallel paths.They consist of :class:`LinearPath` objects.
    """

    def __init__(self, *paths: LinearPath | PathSequence):
        self.paths = list(paths)
        for p in self.paths:
            assert isinstance(p, LinearPath | PathSequence), \
                "Parallel paths must consist of LinearPath or PathSequence."

    def __iter__(self) -> Iterator[LinearPath | PathSequence]:
        return iter(self.paths)

    def __len__(self) -> int:
        return len(self.paths)

    def __getitem__(self, index: int) -> LinearPath | PathSequence:
        return self.paths[index]

    def __bool__(self) -> bool:
        return any(bool(path) for path in self)

    @property
    def input_permutation(self) -> Permutation | None:
        """
        The permutation of the input to the first module.

        If all input_permutations of the paths (except for empty paths)
        are the same, it returns the common permutation (even if it is :code:`None`),
        else it returns :code:`None`.
        """
        perms = [path.input_permutation for path in self if bool(path)]
        if not all(perm is perms[0] for perm in perms):
            return None
        return perms[0]

    @input_permutation.setter
    def input_permutation(self, permutation: Permutation | None) -> None:
        if permutation is None:
            self._set_all_input_permutations(permutation)
            return

        shapes = [path.input_permutation_shape for path in self if bool(path)]
        if (
                permutation.perm_indices.shape[0] == shapes[0]
                and all(shape == shapes[0] for shape in shapes)
        ):
            self._set_all_input_permutations(permutation)

    def _set_all_input_permutations(self, permutation: Permutation | None) -> None:
        for path in self:
            path.input_permutation = permutation

    @property
    def input_permutation_shape(self) -> int:
        shapes = [path.input_permutation_shape for path in self if bool(path)]
        if not all(shape == shapes[0] for shape in shapes):
            return 0
        return shapes[0]

    @property
    def input_shape(self) -> list[tuple[int, ...]]:
        shapes = [path.input_shape for path in self if bool(path)]
        if not all(shape == shapes[0] for shape in shapes):
            return []
        return shapes[0]

    @property
    def output_permutation(self) -> Permutation | None:
        """The permutation of the output of the last module."""
        perms = [path.output_permutation for path in self if bool(path)]
        if not all(perm is perms[0] for perm in perms):
            return None
        return perms[0]

    @output_permutation.setter
    def output_permutation(self, permutation: Permutation) -> None:
        if permutation is None:
            self._set_all_output_permutations(permutation)
            return

        shapes = [path.output_permutation_shape for path in self if bool(path)]
        if (
                permutation.perm_indices.shape[0] == shapes[0]
                and all(shape == shapes[0] for shape in shapes)
        ):
            self._set_all_output_permutations(permutation)

    def _set_all_output_permutations(self, permutation: Permutation | None) -> None:
        for path in self:
            path.output_permutation = permutation

    @property
    def output_permutation_shape(self) -> int:
        shapes = [path.output_permutation_shape for path in self if bool(path)]
        if not all(shape == shapes[0] for shape in shapes):
            return 0
        return shapes[0]

    @property
    def output_shape(self) -> list[tuple[int, ...]]:
        shapes = [path.output_shape for path in self if bool(path)]
        if not all(shape == shapes[0] for shape in shapes):
            return []
        return shapes[0]

    @property
    def permutation_to_info(self) -> list[tuple[Permutation, list[PermutationInfo]]]:
        id_to_perminfo: dict[int, tuple[Permutation, list[PermutationInfo]]] = {}

        for path in self:
            for permutation, info in path.permutation_to_info:
                if permutation is None:
                    continue
                if id(permutation) not in id_to_perminfo:
                    id_to_perminfo[id(permutation)] = (permutation, info)
                id_to_perminfo[id(permutation)][1].extend(info)

        return [perm_info for perm_info in id_to_perminfo.values()]

    def enforce_identity(
            self,
            prev_path: LinearPath | ParallelPaths | None = None,
            next_path: LinearPath | ParallelPaths | None = None
    ) -> None:
        """
        Enforce the constraint that the permutations in :class:`ParallelPaths`
        represent the identity function in sum.

        In other words, applying the permutations in :class:`ParallelPaths`
        changes the layout of the weights and biases of the model, but not
        its output.
        """
        for path in self:
            path.enforce_identity(prev_path=prev_path, next_path=next_path)

        # So far: input_permutation is None if prev_path is None,
        #   output_permutation is None if next_path is None.
        # Also, input_permutation is the previous output_permutation
        #   if prev_path is not None,
        #   output_permutation is the next input_permutation if next_path is not None.
        # Next, handle the edge-cases:
        #
        # 1. If both prev_path and next_path are None,
        #       then we are done (input_permutation and output_permutation are None).
        # 2. If there is no empty path in self,
        #       then self.output_permutation has to be synchronized
        #       between the paths in self.
        #       The may be short paths where the output_permutation
        #       is the input_permutation, which is None,
        #       in which case all output_permutations have to be set to None
        #       Otherwise, either all output_permutations are None,
        #       in which case setting all to the output_permutation of the first path
        #       works fine, or none of them are None, in which case it is fine
        #       to set all output_permutations to any one of the output permutations,
        #       as long as they are all the same.
        # 3. If there is an empty path in self, and prev_path is None,
        #       then output_permutation should also be set to None.
        # 4. If there is an empty path in self, and next_path is None,
        #       then input_permutation needs to be None,
        #       and so does prev_path.output_permutation.
        # 5. If there is an empty path in self,
        #       and neither prev_path nor next_path are None,
        #       then self.output_permutation has to be prev_path.output_permutation.

        # 1. Done
        if prev_path is None and next_path is None:
            return

        # No empty path in self
        if all(bool(path) for path in self):
            # 2. Synchronize output_permutation
            if any(path.output_permutation is None for path in self):
                self.output_permutation = None
            else:
                self.output_permutation = self[0].output_permutation
            return

        # Empty path in self
        # 3. Set output_permutation to None
        if prev_path is None:
            self.output_permutation = None
            return

        # 4. Set input_permutation and prev_path.output_permutation to None
        if next_path is None:
            self.input_permutation = None
            prev_path.output_permutation = None
            return

        # 5. Synchronize output_permutation
        self.output_permutation = prev_path.output_permutation

    def apply_permutations(self) -> None:
        """Apply the permutations in the path to the model."""
        for path in self:
            path.apply_permutations()

    def __repr__(self) -> str:
        path_strings = [repr(path) for path in self]
        max_len = max(
            len(path_string.splitlines()) for path_string in path_strings)

        for i, path_string in enumerate(path_strings):
            # Pad the path string with spaces
            #   so that the parallel paths don't touch.
            path_strings[i] = "\n".join(
                line + " "*4
                for line in path_string.splitlines()
            )

            # Add "|"-lines to the bottom of the path_string
            #   to make it as long as the longest path.
            # The "|" should be padded by spaces such that it is centered
            #   in a line of length max_width.
            max_width = max(len(line) for line in path_string.splitlines())
            centered_line = (
                    "\n"
                    + " " * (max_width // 2 - 1)
                    + "|"
                    + " " * (max_width // 2 + 3)  # adjust for padding on the right
            )
            path_strings[i] += centered_line * (max_len - len(path_string.splitlines()))

        # Join the path strings line-by-line
        final_path_strings = ["  " for _ in range(max_len)]
        for path_string in path_strings:
            for i, line in enumerate(path_string.splitlines()):
                final_path_strings[i] += line

        return "ParallelPaths(" + "\n".join(final_path_strings) + "\n)"


class PathSequence:
    """
    The graph of the model.

    Consists of :class:`LinearPath` and :class:`ParallelPaths`.
    """
    def __init__(self, *paths: LinearPath | ParallelPaths) -> None:
        self.paths = list(paths)
        assert all(
            isinstance(path, LinearPath | ParallelPaths)
            for path in self.paths
        ), "paths must be instances of LinearPath or ParallelPaths"

    def __iter__(self) -> Iterator[LinearPath | ParallelPaths]:
        return iter(self.paths)

    def __len__(self) -> int:
        return len(self.paths)

    def __getitem__(self, index: int) -> LinearPath | ParallelPaths:
        return self.paths[index]

    @property
    def input_permutation(self) -> Permutation | None:
        return self[0].input_permutation

    @input_permutation.setter
    def input_permutation(self, permutation: Permutation | None) -> None:
        self[0].input_permutation = permutation

    @property
    def input_permutation_shape(self) -> int:
        return self[0].input_permutation_shape

    @property
    def input_shape(self) -> list[tuple[int, ...]]:
        return self[0].input_shape

    @property
    def output_permutation(self) -> Permutation | None:
        return self[-1].output_permutation

    @output_permutation.setter
    def output_permutation(self, permutation: Permutation | None) -> None:
        self[-1].output_permutation = permutation

    @property
    def output_permutation_shape(self) -> int:
        return self[-1].output_permutation_shape

    @property
    def output_shape(self) -> list[tuple[int, ...]]:
        return self[-1].output_shape

    @property
    def permutation_to_info(self) -> list[tuple[Permutation, list[PermutationInfo]]]:
        id_to_perminfo: dict[int, tuple[Permutation, list[PermutationInfo]]] = {}

        for path in self:
            for permutation, info in path.permutation_to_info:
                if permutation is None:
                    continue
                if id(permutation) not in id_to_perminfo:
                    id_to_perminfo[id(permutation)] = (permutation, info)
                id_to_perminfo[id(permutation)][1].extend(info)

        return [perm_info for perm_info in id_to_perminfo.values()]

    def enforce_identity(
            self,
            prev_path: LinearPath | ParallelPaths | None = None,
            next_path: LinearPath | ParallelPaths | None = None
    ) -> None:
        """Enforce the identity constraint on all paths."""
        for i in range(len(self)):
            prev_path_ = self[i - 1] if i > 0 else prev_path
            next_path_ = self[i + 1] if i < len(self) - 1 else next_path
            self[i].enforce_identity(prev_path=prev_path_, next_path=next_path_)

    def apply_permutations(self) -> None:
        """Apply the permutations in the paths to the model."""
        for path in self:
            path.apply_permutations()

    def __repr__(self) -> str:
        reprs: list[str] = []
        for i, path in enumerate(self):
            path_repr = repr(path)
            width = max(len(line) for line in path_repr.splitlines())
            if i > 0:
                reprs.append(" " * (width // 2 - 1) + "|" + " " * (width // 2 - 1))
                reprs.append(" " * (width // 2 - 1) + "|" + " " * (width // 2 - 1))
                reprs.append(" " * (width // 2 - 1) + "|" + " " * (width // 2 - 1))
            reprs.append("-" * width)
            reprs.append(path_repr)
            reprs.append("-" * width)

        reprs = ["\nPathSequence(", *reprs] + [")"]
        width = max(len(line) for text in reprs for line in text.splitlines())

        for i, text in enumerate(reprs):
            reprs[i] = "\n".join(
                line + " " * (width - len(line))
                for line in text.splitlines()
            )

        return "\n".join(reprs)



# type: ignore
# I do type-checks in code, but mypy doesn't understand that.
# This leads to me getting an error from mypy at almost every line,
# which is annoying.
# Instead of reformatting my file such that mypy doesn't complain,
# I just ignore all the errors and make sure to test thoroughly.

from __future__ import annotations

import copy

import pytest
import torch
from torch import nn

from rebasin.modules import (
    DefaultModule,
    LayerNormModule,
    ModuleBase,
    MultiheadAttentionModule,
    OneDimModule,
    Permutation,
    PermutationInfo,
    initialize_module,
)
from tests.fixtures.utils import model_change_percent, modules_and_module_nodes


class TestPermutation:
    @staticmethod
    def test_len() -> None:
        permutation = Permutation(torch.randperm(5))
        assert len(permutation) == len(permutation.perm_indices) == 5

    @staticmethod
    def test_eq() -> None:
        permutation = Permutation(torch.tensor([0, 2, 1, 3]))
        assert permutation == Permutation(torch.tensor([0, 2, 1, 3]))
        assert permutation != Permutation(torch.tensor([0, 1, 2, 3]))
        assert permutation != Permutation(torch.tensor([0, 2, 1, 3, 4]))
        assert permutation != "not a Permutation"


class TestPermutationInfo:
    @staticmethod
    def test_eq() -> None:
        lin_a, lin_b, node = modules_and_module_nodes(
            nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
        )
        module = ModuleBase(lin_a, lin_b, node)
        permutation_info = PermutationInfo(
            module=module,
            axis=0,
            parameter_a=lin_a.weight,
            parameter_b=lin_b.weight,
        )
        assert permutation_info == PermutationInfo(
            module=module,
            axis=0,
            parameter_a=lin_a.weight,
            parameter_b=lin_b.weight,
        )
        assert permutation_info != PermutationInfo(
            module=module,
            axis=1,
            parameter_a=lin_a.weight,
            parameter_b=lin_b.weight,
        )
        assert permutation_info != PermutationInfo(
            module=module,
            axis=0,
            parameter_a=lin_a.bias,
            parameter_b=lin_b.weight,
        )
        assert permutation_info != PermutationInfo(
            module=module,
            axis=0,
            parameter_a=lin_a.weight,
            parameter_b=lin_b.bias,
        )
        assert permutation_info != "not a PermutationInfo"


def test_base_module() -> None:
    with pytest.raises(TypeError):
        ModuleBase(
            *modules_and_module_nodes(
                nn.Linear(5, 5), nn.Conv2d(5, 5, 3), torch.randn(1, 5, 15, 15)
            )
        )

    with pytest.raises(TypeError):
        _, _, node = modules_and_module_nodes(
            nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
        )
        ModuleBase("foo", "bar", node)

    mb = ModuleBase(
        *modules_and_module_nodes(
            nn.Conv2d(5, 5, 3), nn.Conv2d(5, 5, 3), torch.randn(1, 5, 15, 15)
        )
    )
    with pytest.raises(NotImplementedError):
        _ = mb.input_permutation

    with pytest.raises(NotImplementedError):
        mb.input_permutation = Permutation(torch.arange(5))

    with pytest.raises(NotImplementedError):
        _ = mb.output_permutation

    with pytest.raises(NotImplementedError):
        mb.output_permutation = Permutation(torch.arange(5))

    with pytest.raises(NotImplementedError):
        _ = mb.permutation_to_info

    with pytest.raises(NotImplementedError):
        mb.apply_permutations()

    with pytest.raises(NotImplementedError):
        _ = mb.input_permutation_shape

    with pytest.raises(NotImplementedError):
        _ = mb.output_permutation_shape


class TestDefaultModule:
    @staticmethod
    def test_sanity_checks() -> None:
        with pytest.raises(AttributeError):
            m1, m2, node = modules_and_module_nodes(
                nn.MultiheadAttention(5, 5),
                nn.MultiheadAttention(5, 5),
                (torch.randn(5, 5), torch.randn(5, 5), torch.randn(5, 5))
            )
            DefaultModule(m1, m2, node)

        with pytest.raises(AttributeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            del lin_b.weight
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(AttributeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            del lin_a.bias
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(AttributeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            del lin_b.bias
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(ValueError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 3), nn.Linear(5, 5), torch.randn(5)
            )
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(ValueError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5, bias=False), torch.randn(5)
            )
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(ValueError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            lin_b.bias = nn.Parameter(torch.zeros(3))
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(TypeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            lin_a.weight = nn.Linear(5, 5)
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(TypeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            lin_b.weight = nn.Linear(5, 5)
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(TypeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            lin_a.bias = nn.Linear(5, 5)
            DefaultModule(lin_a, lin_b, node)

        with pytest.raises(TypeError):
            lin_a, lin_b, node = modules_and_module_nodes(
                nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5)
            )
            lin_b.bias = nn.Linear(5, 5)
            DefaultModule(lin_a, lin_b, node)

    @staticmethod
    def test_io_permutations() -> None:
        mb = initialize_module(
            *modules_and_module_nodes(nn.Linear(3, 5), nn.Linear(3, 5), torch.randn(3))
        )
        assert mb.input_permutation == Permutation(torch.arange(3))
        assert mb.output_permutation == Permutation(torch.arange(5))

        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        assert mb.input_permutation == Permutation(torch.tensor([2, 0, 1]))

        mb.output_permutation = Permutation(torch.tensor([4, 2, 3, 1, 0]))
        assert mb.output_permutation == Permutation(torch.tensor([4, 2, 3, 1, 0]))

    @staticmethod
    def test_permutation_to_info() -> None:
        mb = initialize_module(
            *modules_and_module_nodes(nn.Linear(3, 5), nn.Linear(3, 5), torch.randn(3))
        )
        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        mb.output_permutation = Permutation(torch.tensor([4, 2, 3, 1, 0]))

        info = mb.permutation_to_info
        assert info == [
            (
                mb.output_permutation,
                [
                    PermutationInfo(mb, 0, mb.module_a.weight, mb.module_b.weight),
                    PermutationInfo(mb, 0, mb.module_a.bias, mb.module_b.bias),
                ]
            ),
            (
                mb.input_permutation,
                [PermutationInfo(mb, 1, mb.module_a.weight, mb.module_b.weight)]
            )
        ]

        mb = initialize_module(
            *modules_and_module_nodes(
                nn.Linear(3, 5, bias=False), nn.Linear(3, 5, bias=False), torch.randn(3)
            )
        )
        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        mb.output_permutation = Permutation(torch.tensor([4, 2, 3, 1, 0]))

        info = mb.permutation_to_info
        assert info == [
            (
                mb.output_permutation,
                [PermutationInfo(mb, 0, mb.module_a.weight, mb.module_b.weight)]
            ),
            (
                mb.input_permutation,
                [PermutationInfo(mb, 1, mb.module_a.weight, mb.module_b.weight)]
            )
        ]

        # What if the input-and output-Permutation is the same?

        mb = initialize_module(
            *modules_and_module_nodes(nn.Linear(5, 5), nn.Linear(5, 5), torch.randn(5))
        )
        mb.input_permutation = Permutation(torch.tensor([2, 0, 1, 3, 4]))
        mb.output_permutation = mb.input_permutation

        info = mb.permutation_to_info
        assert info == [
            (
                mb.input_permutation,
                [
                    PermutationInfo(mb, 0, mb.module_a.weight, mb.module_b.weight),
                    PermutationInfo(mb, 0, mb.module_a.bias, mb.module_b.bias),
                    PermutationInfo(mb, 1, mb.module_a.weight, mb.module_b.weight),
                ]
            )
        ]

    @staticmethod
    def test_apply_permutations() -> None:
        lin_a, lin_b, node = modules_and_module_nodes(
            nn.Linear(3, 5), nn.Linear(3, 5), torch.randn(3)
        )
        lin_b_orig = copy.deepcopy(lin_b)
        mb = initialize_module(lin_a, lin_b, node)

        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        mb.output_permutation = Permutation(torch.tensor([4, 2, 3, 1, 0]))
        mb.apply_permutations(except_axis=0)

        assert not torch.allclose(lin_b.weight, lin_b_orig.weight)
        assert torch.allclose(
            lin_b.weight[:, torch.argsort(mb.input_permutation.perm_indices)],
            lin_b_orig.weight
        )
        assert torch.allclose(lin_b.bias, lin_b_orig.bias)

        mb.apply_permutations(except_axis=1)
        assert not torch.allclose(lin_b.bias, lin_b_orig.bias)

        # Does except_axis=-1 mean that permutations are applied to all axes?
        mb.input_permutation = Permutation(
            torch.argsort(mb.input_permutation.perm_indices)
        )
        mb.output_permutation = Permutation(
            torch.argsort(mb.output_permutation.perm_indices)
        )

        mb.apply_permutations()
        assert torch.allclose(lin_b.weight, lin_b_orig.weight)
        assert torch.allclose(lin_b.bias, lin_b_orig.bias)

    @staticmethod
    def test_permute_axis() -> None:
        parameter = nn.Parameter(torch.randn(3, 3))
        parameter_orig = copy.deepcopy(parameter)

        permutation = torch.tensor([2, 0, 1])
        DefaultModule.permute_parameter(parameter, 0, permutation)

        assert not torch.allclose(parameter, parameter_orig)
        assert torch.allclose(
            parameter[torch.argsort(permutation)],
            parameter_orig
        )

    @staticmethod
    def test_none_permuation() -> None:
        lin_a, lin_b, node = modules_and_module_nodes(
            nn.Linear(3, 5), nn.Linear(3, 5), torch.randn(3)
        )
        lin_b_orig = copy.deepcopy(lin_b)
        mb = initialize_module(lin_a, lin_b, node)
        mb.input_permutation = None
        mb.output_permutation = None

        mb.apply_permutations()
        assert torch.allclose(lin_b.weight, lin_b_orig.weight)

        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        mb.apply_permutations()
        assert torch.allclose(
            lin_b.weight[:, torch.argsort(mb.input_permutation.perm_indices)],
            lin_b_orig.weight
        )

        mb.input_permutation = Permutation(
            torch.argsort(mb.input_permutation.perm_indices)
        )
        mb.output_permutation = Permutation(torch.tensor([4, 2, 3, 1, 0]))
        mb.apply_permutations()

        assert torch.allclose(
            lin_b.weight[torch.argsort(mb.output_permutation.perm_indices)],
            lin_b_orig.weight
        )

    @staticmethod
    def test_shapes() -> None:
        lin_a, lin_b, node = modules_and_module_nodes(
            nn.Linear(3, 5), nn.Linear(3, 5), torch.randn(3)
        )
        mb = initialize_module(lin_a, lin_b, node)
        assert mb.input_permutation_shape == lin_a.weight.shape[1]
        assert mb.output_permutation_shape == lin_a.weight.shape[0]


class TestOneDimModule:
    @staticmethod
    def test_io_permutations() -> None:
        mb = initialize_module(
            *modules_and_module_nodes(
                nn.BatchNorm2d(3), nn.BatchNorm2d(3), torch.randn(1, 3, 12, 12)
            )
        )
        assert isinstance(mb, OneDimModule)
        assert mb.input_permutation == Permutation(torch.arange(3))
        assert mb.input_permutation is mb.output_permutation

        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        assert mb.input_permutation == Permutation(torch.tensor([2, 0, 1]))

    @staticmethod
    def test_apply_permutations() -> None:
        bn_a, bn_b, node = modules_and_module_nodes(
            nn.BatchNorm2d(3), nn.BatchNorm2d(3), torch.randn(1, 3, 12, 12)
        )
        bn_b.weight.data = torch.randn_like(bn_b.weight.data)
        bn_b.bias.data = torch.randn_like(bn_b.bias.data)
        bn_b_orig = copy.deepcopy(bn_b)

        mb = initialize_module(bn_a, bn_b, node)

        mb.input_permutation = Permutation(torch.tensor([2, 0, 1]))
        assert mb.output_permutation is mb.input_permutation
        mb.apply_permutations()

        assert not torch.allclose(bn_b.weight, bn_b_orig.weight)
        assert not torch.allclose(bn_b.bias, bn_b_orig.bias)

    @staticmethod
    def test_none_permutation() -> None:
        mb = initialize_module(
            *modules_and_module_nodes(
                nn.BatchNorm2d(3), nn.BatchNorm2d(3), torch.randn(1, 3, 12, 12)
            )
        )
        assert isinstance(mb, OneDimModule)
        assert mb.input_permutation == Permutation(torch.arange(3))
        assert mb.input_permutation is mb.output_permutation

        mb.input_permutation = None
        assert mb.input_permutation is None
        assert mb.output_permutation is None

    @staticmethod
    def test_shapes() -> None:
        bn_a, bn_b, node = modules_and_module_nodes(
            nn.BatchNorm2d(3), nn.BatchNorm2d(3), torch.randn(1, 3, 12, 12)
        )
        mb = initialize_module(bn_a, bn_b, node)
        assert mb.input_permutation_shape == bn_a.weight.shape[0]
        assert mb.output_permutation_shape == bn_a.weight.shape[0]


class TestLayerNormModule:
    @staticmethod
    def test_permutation_to_info() -> None:
        mb = initialize_module(
            *modules_and_module_nodes(
                nn.LayerNorm([4, 5]), nn.LayerNorm([4, 5]), torch.randn(4, 5)
            )
                     )
        assert isinstance(mb, LayerNormModule)

        mb.input_permutation = Permutation(torch.tensor([2, 0, 1, 3]))
        assert mb.input_permutation == Permutation(torch.tensor([2, 0, 1, 3]))

        mb.output_permutation = Permutation(torch.tensor([2, 0, 1, 3, 4]))
        assert mb.output_permutation == Permutation(torch.tensor([2, 0, 1, 3, 4]))

    @staticmethod
    def test_apply_permutations() -> None:
        ln_a, ln_b, node = modules_and_module_nodes(
            nn.LayerNorm([4, 5]), nn.LayerNorm([4, 5]), torch.randn(4, 5)
        )
        ln_b.weight.data = torch.randn_like(ln_b.weight.data)
        ln_b.bias.data = torch.randn_like(ln_b.bias.data)
        ln_b_orig = copy.deepcopy(ln_b)
        mb = initialize_module(ln_a, ln_b, node)

        mb.input_permutation = Permutation(torch.tensor([2, 0, 4, 1, 3]))
        assert mb.output_permutation is mb.input_permutation
        mb.apply_permutations()

        assert not torch.allclose(ln_b.weight, ln_b_orig.weight)
        assert not torch.allclose(ln_b.bias, ln_b_orig.bias)

    @staticmethod
    def test_none_permutation() -> None:
        mb = initialize_module(
            *modules_and_module_nodes(
                nn.LayerNorm([4, 5]), nn.LayerNorm([4, 5]), torch.randn(4, 5)
            )
        )
        assert isinstance(mb, LayerNormModule)

        mb.input_permutation = None
        assert mb.input_permutation is None
        assert mb.output_permutation is None

    @staticmethod
    def test_shapes() -> None:
        ln_a, ln_b, node = modules_and_module_nodes(
            nn.LayerNorm([4, 5]), nn.LayerNorm([4, 5]), torch.randn(4, 5)
        )
        mb = initialize_module(ln_a, ln_b, node)
        assert mb.input_permutation_shape == ln_a.weight.shape[1]
        assert mb.output_permutation_shape == ln_a.weight.shape[1]

        ln_a, ln_b, node = modules_and_module_nodes(
            nn.LayerNorm(3), nn.LayerNorm(3), torch.randn(4, 3)
        )
        mb = initialize_module(ln_a, ln_b, node)
        assert mb.input_permutation_shape == ln_a.weight.shape[0]
        assert mb.output_permutation_shape == ln_a.weight.shape[0]


class TestMultiheadAttentionModule:
    @staticmethod
    def test_permutation_to_info() -> None:
        embed_dim = 6
        num_heads = 3
        x = torch.randn(6, 6, 6)
        mha, mhb, node = modules_and_module_nodes(
            nn.MultiheadAttention(embed_dim, num_heads),
            nn.MultiheadAttention(embed_dim, num_heads),
            (x, x, x),
        )

        mb = MultiheadAttentionModule(mha, mhb, node)
        mb.input_permutation = Permutation(torch.tensor([2, 0, 1, 3, 4, 5]))
        mb.output_permutation = Permutation(torch.tensor([4, 2, 3, 1, 0, 5]))

        info = mb.permutation_to_info
        assert info == [
            (
                mb.input_permutation,
                [PermutationInfo(
                    mb, 1, mb.module_a.in_proj_weight, mb.module_b.in_proj_weight
                )]
            ),
            (
                mb.output_permutation,
                [
                    PermutationInfo(
                        mb, 0,
                        mb.module_a.out_proj.weight,
                        mb.module_b.out_proj.weight
                    ),
                    PermutationInfo(
                        mb, 0,
                        mb.module_a.out_proj.bias,
                        mb.module_b.out_proj.bias
                    ),
                ]
            ),
        ]

    @staticmethod
    def test_input_permutation() -> None:
        embed_dim = 6
        num_heads = 3
        x = torch.randn(6, 6)
        mha, mhb, node = modules_and_module_nodes(
            nn.MultiheadAttention(embed_dim, num_heads, bias=False),
            nn.MultiheadAttention(embed_dim, num_heads, bias=False),
            (x, x, x),
        )
        mhb_orig = copy.deepcopy(mhb)

        mb = MultiheadAttentionModule(mha, mhb, node)
        perm = Permutation(torch.tensor([2, 0, 1, 3, 4, 5]))
        mb.input_permutation = perm
        mb.output_permutation = perm

        assert mb.input_permutation == perm
        assert mb.output_permutation == perm
        mb.apply_permutations()
        assert model_change_percent(mhb_orig, mhb) > 0.1

        mha, mhb, node = modules_and_module_nodes(
            nn.MultiheadAttention(embed_dim, num_heads, bias=False, kdim=4, vdim=5),
            nn.MultiheadAttention(embed_dim, num_heads, bias=False, kdim=4, vdim=5),
            (x, torch.randn(6, 4), torch.randn(6, 5)),
        )
        mhb_orig = copy.deepcopy(mhb)
        mb = MultiheadAttentionModule(mha, mhb, node)
        perm = Permutation(torch.tensor([2, 0, 1, 3, 4, 5]))
        mb.input_permutation = perm
        mb.output_permutation = None

        assert mb.input_permutation is None
        assert mb.output_permutation is None
        mb.apply_permutations()

        assert model_change_percent(mhb_orig, mhb) < 1e-6

    @staticmethod
    def test_io() -> None:
        embed_dim = 6
        num_heads = 3

        class Model(nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.mha = nn.MultiheadAttention(embed_dim, num_heads)
                self.lin1 = nn.Linear(embed_dim, embed_dim)
                self.lin2 = nn.Linear(embed_dim, embed_dim)
                self.relu = nn.ReLU()

            def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:
                input_tensor = self.relu(self.lin1(input_tensor))
                input_tensor = self.relu(
                    self.mha(input_tensor, input_tensor, input_tensor)[0]
                )
                input_tensor = self.relu(self.lin2(input_tensor))
                return input_tensor

        model = Model()
        model_orig = copy.deepcopy(model)

        x = torch.randn(embed_dim, embed_dim)
        y_orig = model(x)

        lin1_mod = initialize_module(
            *modules_and_module_nodes(model.lin1, model.lin1, x)
        )
        mha_mod = initialize_module(
            *modules_and_module_nodes(
                model.mha, model.mha, (model.lin1(x), model.lin1(x), model.lin1(x))
            )
        )
        lin2_mod = initialize_module(
            *modules_and_module_nodes(model.lin2, model.lin2, x)
        )

        perm1 = Permutation(torch.tensor([2, 0, 1, 3, 4, 5]))
        perm2 = Permutation(torch.tensor([0, 4, 2, 5, 1, 3]))

        lin1_mod.output_permutation = perm1
        mha_mod.input_permutation = perm1
        mha_mod.output_permutation = perm2
        lin2_mod.input_permutation = perm2

        lin1_mod.apply_permutations()
        mha_mod.apply_permutations()
        lin2_mod.apply_permutations()

        assert model_change_percent(model_orig, model) > 0.1

        y_new = model(x)
        assert torch.allclose(y_orig, y_new)

    @staticmethod
    def test_shapes() -> None:
        embed_dim = 6
        num_heads = 3
        x = torch.randn(6, 6)
        mha, mhb, node = modules_and_module_nodes(
            nn.MultiheadAttention(embed_dim, num_heads),
            nn.MultiheadAttention(embed_dim, num_heads),
            (x, x, x),
        )
        mb = initialize_module(mha, mhb, node)
        assert isinstance(mhb.in_proj_weight, nn.Parameter)
        assert mb.input_permutation_shape == embed_dim
        assert mb.output_permutation_shape == embed_dim

        mha, mhb, node = modules_and_module_nodes(
            nn.MultiheadAttention(embed_dim, num_heads, kdim=4, vdim=5),
            nn.MultiheadAttention(embed_dim, num_heads, kdim=4, vdim=5),
            (x, torch.randn(6, 4), torch.randn(6, 5)),
        )
        mb = initialize_module(mha, mhb, node)
        assert mhb.in_proj_weight is None
        assert mb.input_permutation_shape == 0
        assert mb.output_permutation_shape == embed_dim



from __future__ import annotations

import copy
import sys
from typing import Any

import pytest
import torch
from torch import nn

from rebasin.modules import (  # type: ignore[attr-defined]
    Permutation,
    initialize_module,
)
from rebasin.paths import LinearPath, ParallelPaths, PathSequence
from tests.fixtures.paths import PathSource
from tests.fixtures.utils import (
    allclose,
    model_change_percent,
    tensor_diff_perc,
)


class TestLinearPath(PathSource):
    def test_len(self) -> None:
        _, path = self.dense_lin_path()
        assert len(path) == 4

    def test_iter(self) -> None:
        _, path = self.dense_lin_path()
        assert list(path) == list(path.modules)

    def test_getitem(self) -> None:
        _, path = self.dense_lin_path()
        assert path[0] == path.modules[0]
        assert path[1] == path.modules[1]

    def test_bool(self) -> None:
        _, path = self.dense_lin_path()
        assert bool(path) is True
        assert not LinearPath()

    def test_input_permutation(self) -> None:
        _, lin_path = self.dense_lin_path()
        assert lin_path.input_permutation == lin_path[0].input_permutation

        permutation = Permutation(torch.randperm(10))
        lin_path.input_permutation = permutation
        assert lin_path.input_permutation is permutation
        assert lin_path[1].input_permutation is permutation  # It's a 1d-module

        _, conv_path = self.conv_lin_path()
        assert conv_path.input_permutation == conv_path[0].input_permutation

        permutation = Permutation(torch.randperm(3))
        conv_path.input_permutation = permutation
        assert conv_path.input_permutation is permutation
        assert conv_path[1].input_permutation is permutation  # It's a 2d-module
        assert conv_path[1].output_permutation is not permutation

    def test_output_permutation(self) -> None:
        _, lin_path = self.dense_lin_path()
        assert lin_path.output_permutation == lin_path[-1].output_permutation

        permutation = Permutation(torch.randperm(10))
        lin_path.output_permutation = permutation
        assert lin_path.output_permutation is permutation
        assert lin_path[-1].input_permutation is permutation  # It's a 1d-module
        assert lin_path[-2].output_permutation is permutation  # It's 1d and a LayerNorm

        _, conv_path = self.conv_lin_path()
        assert conv_path.output_permutation == conv_path[-1].output_permutation

        permutation = Permutation(torch.randperm(3))
        conv_path.output_permutation = permutation
        assert conv_path.output_permutation is permutation
        assert conv_path[-1].output_permutation is permutation  # It's a LayerNorm
        assert conv_path[-1].input_permutation is permutation
        assert conv_path[-2].output_permutation is permutation

    def test_io_linear(self) -> None:
        lin_model, lin_path = self.dense_lin_path()
        lin_model_orig = copy.deepcopy(lin_model)
        x = torch.randn(10, 10)
        y_orig = lin_model(x)

        for permutation, _ in lin_path.permutation_to_info:
            permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

        lin_path.input_permutation = None
        lin_path.output_permutation = None

        lin_path.enforce_identity()
        lin_path.apply_permutations()
        assert model_change_percent(lin_model, lin_model_orig) > 0.1

        y_new = lin_model(x)
        assert allclose(y_orig, y_new)

    def test_io_conv(self) -> None:
        conv_model, conv_path = self.conv_lin_path()
        conv_model_orig = copy.deepcopy(conv_model)
        x = torch.randn(1, 3, 10, 10)
        y_orig = conv_model(x)

        for permutation, _ in conv_path.permutation_to_info:
            while torch.all(
                    permutation.perm_indices
                    == torch.arange(len(permutation.perm_indices))
            ):
                permutation.perm_indices = torch.randperm(len(permutation.perm_indices))

        conv_path.input_permutation = None
        conv_path.output_permutation = None

        conv_path.enforce_identity()
        conv_path.apply_permutations()
        assert model_change_percent(conv_model, conv_model_orig) > 0.02

        y_new = conv_model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_lin_path()
        model3, path3 = self.dense_lin_path()

        model = nn.Sequential(model1, model2, model3)
        model_orig = copy.deepcopy(model)
        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for (perm1, _), (perm2, _), (perm3, _) in zip(
                path1.permutation_to_info,
                path2.permutation_to_info,
                path3.permutation_to_info
        ):
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))
            perm3.perm_indices = torch.randperm(len(perm3.perm_indices))

        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=path3)
        path3.enforce_identity(prev_path=path2, next_path=None)

        assert path1.input_permutation is None
        assert isinstance(path1.output_permutation, Permutation)
        assert isinstance(path2.input_permutation, Permutation)
        assert isinstance(path2.output_permutation, Permutation)
        assert isinstance(path3.input_permutation, Permutation)
        assert path3.output_permutation is None

        path1.apply_permutations()
        path2.apply_permutations()
        path3.apply_permutations()
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_shapes(self) -> None:
        model, path = self.dense_lin_path()
        assert path.input_permutation_shape == 10
        assert path.output_permutation_shape == 10

        model, path = self.conv_lin_path()
        assert path.input_permutation_shape == \
               model[0].weight.shape[1]  # type: ignore[index]
        assert path.output_permutation_shape == \
               model[-1].weight.shape[1]  # type: ignore[index]

    @staticmethod
    def test_empty_path() -> None:
        path = LinearPath()
        assert len(path) == 0
        assert not path
        assert not path.modules
        assert not path.permutation_to_info
        assert path.input_permutation is None
        assert path.output_permutation is None
        path.input_permutation = torch.randperm(4)
        assert path.input_permutation is None
        path.output_permutation = torch.randperm(4)
        assert path.output_permutation is None
        path.enforce_identity()
        path.apply_permutations()


class TestParallelPaths(PathSource):
    def test_iter(self) -> None:
        _, path = self.dense_parallel_path_with_empty_path()
        for linpath in path:
            assert isinstance(linpath, LinearPath)

    def test_len(self) -> None:
        _, path = self.dense_parallel_path_with_empty_path()
        assert len(path) == 3

    def test_getitem(self) -> None:
        _, path = self.dense_parallel_path_with_empty_path()
        assert path[0] == path.paths[0]
        assert path[1] == path.paths[1]
        assert path[2] == path.paths[2]

    def test_input_permutation(self) -> None:
        _, path = self.dense_parallel_path_with_empty_path()
        assert path.input_permutation is None
        path.input_permutation = Permutation(torch.randperm(10))
        assert isinstance(path.input_permutation, Permutation)
        path.input_permutation = None
        assert path.input_permutation is None

        _, path = self.dense_parallel_path_diff_shapes()
        assert path.input_permutation is None
        path.input_permutation = Permutation(torch.randperm(10))

        assert path.input_permutation is None

    def test_output_permutation(self) -> None:
        _, path = self.dense_parallel_path_with_empty_path()
        assert path.output_permutation is None
        path.output_permutation = Permutation(torch.randperm(10))
        assert isinstance(path.output_permutation, Permutation)
        path.output_permutation = None
        assert path.output_permutation is None

        _, path = self.dense_parallel_path_diff_shapes()
        assert path.output_permutation is None
        path.output_permutation = Permutation(torch.randperm(10))
        assert path.output_permutation is None

    def test_enforce_identity_between_paths_with_empty_path(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_parallel_path_with_empty_path()
        model3, path3 = self.dense_lin_path()

        model = nn.Sequential(model1, model2, model3)
        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model3_orig = copy.deepcopy(model3)
        model_orig = copy.deepcopy(model)
        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for (perm1, _), (perm2, _), (perm3, _) in zip(
                path1.permutation_to_info,
                path2.permutation_to_info,
                path3.permutation_to_info
        ):
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))
            perm3.perm_indices = torch.randperm(len(perm3.perm_indices))

        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=path3)
        path3.enforce_identity(prev_path=path2, next_path=None)

        path1.apply_permutations()
        path2.apply_permutations()
        path3.apply_permutations()
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model3, model3_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_between_paths_no_empty_path(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_parallel_path_no_empty_path()
        model3, path3 = self.dense_lin_path()

        model = nn.Sequential(model1, model2, model3)
        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model3_orig = copy.deepcopy(model3)
        model_orig = copy.deepcopy(model)
        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for (perm1, _), (perm2, _), (perm3, _) in zip(
                path1.permutation_to_info,
                path2.permutation_to_info,
                path3.permutation_to_info
        ):
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))
            perm3.perm_indices = torch.randperm(len(perm3.perm_indices))

        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=path3)
        path3.enforce_identity(prev_path=path2, next_path=None)

        path1.apply_permutations()
        path2.apply_permutations()
        path3.apply_permutations()

        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model3, model3_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_prev_path_none_with_empty_path(self) -> None:
        model1, path1 = self.dense_parallel_path_with_empty_path()
        model2, path2 = self.dense_lin_path()

        model = nn.Sequential(model1, model2)
        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model_orig = copy.deepcopy(model)
        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for (perm1, _), (perm2, _) in zip(
                path1.permutation_to_info,
                path2.permutation_to_info
        ):
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))

        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=None)

        path1.apply_permutations()
        path2.apply_permutations()

        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_prev_path_none_no_empty_path(self) -> None:
        # Setup
        model1, path1 = self.dense_parallel_path_no_empty_path()
        model2, path2 = self.dense_lin_path()
        model = nn.Sequential(model1, model2)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm1, _ in path1.permutation_to_info:
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))

        for perm2, _ in path2.permutation_to_info:
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))

        # Permute
        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=None)

        path1.apply_permutations()
        path2.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_next_path_none_with_empty_path(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_parallel_path_with_empty_path()
        model = nn.Sequential(model1, model2)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm1, _ in path1.permutation_to_info:
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))

        for perm2, _ in path2.permutation_to_info:
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))

        # Permute
        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=None)

        path1.apply_permutations()
        path2.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_next_path_none_no_empty_path(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_parallel_path_no_empty_path()
        model = nn.Sequential(model1, model2)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm1, _ in path1.permutation_to_info:
            perm1.perm_indices = torch.randperm(len(perm1.perm_indices))

        for perm2, _ in path2.permutation_to_info:
            perm2.perm_indices = torch.randperm(len(perm2.perm_indices))

        # Permute
        path1.enforce_identity(prev_path=None, next_path=path2)
        path2.enforce_identity(prev_path=path1, next_path=None)

        path1.apply_permutations()
        path2.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_no_other_paths_with_empty_path(self) -> None:
        model, path = self.dense_parallel_path_with_empty_path()
        model_orig = copy.deepcopy(model)
        x = torch.randn(10, 10)
        y_orig = model(x)

        for perm, _ in path.permutation_to_info:
            perm.perm_indices = torch.randperm(len(perm.perm_indices))

        path.enforce_identity(prev_path=None, next_path=None)
        path.apply_permutations()
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_enforce_identity_no_other_paths_no_empty_path(self) -> None:
        model, path = self.dense_parallel_path_no_empty_path()
        model_orig = copy.deepcopy(model)
        x = torch.randn(10, 10)
        y_orig = model(x)

        for perm, _ in path.permutation_to_info:
            perm.perm_indices = torch.randperm(len(perm.perm_indices))

        path.enforce_identity(prev_path=None, next_path=None)
        assert path.input_permutation is None
        assert path.output_permutation is None

        path.apply_permutations()
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)


class TestModelPath(PathSource):
    def test_lin_par_lin_no_empty_path(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_parallel_path_no_empty_path()
        model3, path3 = self.dense_lin_path()
        model = nn.Sequential(model1, model2, model3)
        graph = PathSequence(path1, path2, path3)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model3_orig = copy.deepcopy(model3)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm, _ in graph.permutation_to_info:
            perm.perm_indices = torch.randperm(len(perm.perm_indices))

        # Permute
        graph.enforce_identity()
        graph.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model3, model3_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_lin_par_lin_with_empty_path(self) -> None:
        model1, path1 = self.dense_lin_path()
        model2, path2 = self.dense_parallel_path_with_empty_path()
        model3, path3 = self.dense_lin_path()
        model = nn.Sequential(model1, model2, model3)
        graph = PathSequence(path1, path2, path3)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model3_orig = copy.deepcopy(model3)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm, _ in graph.permutation_to_info:
            perm.perm_indices = torch.randperm(len(perm.perm_indices))

        # Permute
        graph.enforce_identity()
        graph.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model3, model3_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_par_par_par_par(self) -> None:
        model1, path1 = self.dense_parallel_path_no_empty_path()
        model2, path2 = self.dense_parallel_path_with_empty_path()
        model3, path3 = self.dense_parallel_path_no_empty_path()
        model4, path4 = self.dense_parallel_path_with_empty_path()
        model = nn.Sequential(model1, model2, model3, model4)
        graph = PathSequence(path1, path2, path3, path4)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model3_orig = copy.deepcopy(model3)
        model4_orig = copy.deepcopy(model4)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm, _ in graph.permutation_to_info:
            perm.perm_indices = torch.randperm(len(perm.perm_indices))

        # Permute
        graph.enforce_identity()
        graph.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model3, model3_orig) > 0.1
        assert model_change_percent(model4, model4_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)

    def test_par_par_par_par2(self) -> None:
        model1, path1 = self.dense_parallel_path_with_empty_path()
        model2, path2 = self.dense_parallel_path_with_empty_path()
        model3, path3 = self.dense_parallel_path_with_empty_path()
        model4, path4 = self.dense_parallel_path_no_empty_path()
        model = nn.Sequential(model1, model2, model3, model4)
        graph = PathSequence(path1, path2, path3, path4)

        model1_orig = copy.deepcopy(model1)
        model2_orig = copy.deepcopy(model2)
        model3_orig = copy.deepcopy(model3)
        model4_orig = copy.deepcopy(model4)
        model_orig = copy.deepcopy(model)

        x = torch.randn(10, 10)
        y_orig = model(x)

        # Randomize permutations
        for perm, _ in graph.permutation_to_info:
            perm.perm_indices = torch.randperm(len(perm.perm_indices))

        # Permute
        graph.enforce_identity()
        graph.apply_permutations()

        # Assertions
        assert model_change_percent(model1, model1_orig) > 0.1
        assert model_change_percent(model2, model2_orig) > 0.1
        assert model_change_percent(model3, model3_orig) > 0.1
        assert model_change_percent(model4, model4_orig) > 0.1
        assert model_change_percent(model, model_orig) > 0.1

        y_new = model(x)
        assert allclose(y_orig, y_new)


@pytest.mark.skipif(
    "--full-suite" not in sys.argv,
    reason="Long test. Run with --full-suite to include."
)
class TestLengtheningModel(PathSource):
    iters = 100

    def test_par_no_empty_path(self) -> None:
        constructors = [self.dense_parallel_path_no_empty_path]
        self.print_lengthening_model(constructors, self.iters)

    def test_par_empty_path(self) -> None:
        constructors = [self.dense_parallel_path_with_empty_path]
        self.print_lengthening_model(constructors, self.iters)

    def test_lin__par_no_empty_path(self) -> None:
        constructors = [self.dense_lin_path, self.dense_parallel_path_no_empty_path]
        self.print_lengthening_model(constructors, self.iters)

    def test_lin__par_empty_path(self) -> None:
        constructors = [self.dense_lin_path, self.dense_parallel_path_with_empty_path]
        self.print_lengthening_model(constructors, self.iters)

    def test_par_no_empty_path__lin(self) -> None:
        constructors = [self.dense_parallel_path_no_empty_path, self.dense_lin_path]
        self.print_lengthening_model(constructors, self.iters)

    def test_par_empty_path__lin(self) -> None:
        constructors = [self.dense_parallel_path_with_empty_path, self.dense_lin_path]
        self.print_lengthening_model(constructors, self.iters)

    def test_par_empty_path__par_no_empty_path(self) -> None:
        constructors = [
            self.dense_parallel_path_with_empty_path,
            self.dense_parallel_path_no_empty_path
        ]
        self.print_lengthening_model(constructors, self.iters)

    @staticmethod
    def print_lengthening_model(constructors: list[Any], iters: int) -> None:
        model = nn.Sequential()
        paths: list[ParallelPaths | LinearPath] = []
        x = torch.randn(10, 10)

        diffs: list[float] = []

        for _ in range(iters):
            for constructor in constructors:
                model_n, path_n = constructor()
                model.append(model_n)
                paths.append(path_n)
            graph = PathSequence(*paths)
            model_orig = copy.deepcopy(model)

            y_orig = model(x)

            # Randomize permutations
            for perm, _ in graph.permutation_to_info:
                perm.perm_indices = torch.randperm(len(perm.perm_indices))

            # Permute
            graph.enforce_identity()
            graph.apply_permutations()

            # Assertions
            assert model_change_percent(model, model_orig) > 0.1

            y_new = model(x)
            diff = tensor_diff_perc(y_new, y_orig)
            diffs.append(diff)
            print(len(model), diff)
            assert allclose(y_orig, y_new)


TestModelPath().test_par_par_par_par()



from __future__ import annotations

from typing import Any, Union

from torch import nn
from torchview import FunctionNode, ModuleNode, TensorNode, draw_graph

from rebasin.modules import (  # type: ignore[attr-defined]
    MODULE_TYPES,
    initialize_module,
)
from rebasin.paths import LinearPath, ParallelPaths, PathSequence
from rebasin.type_definitions import NODE_TYPES


class PermutationInitializer:
    def __init__(
            self,
            model_a: nn.Module,
            model_b: nn.Module,
            input_data_b: Any,
            input_data_a: Any | None = None,
    ) -> None:
        self.model_a = model_a
        self.model_b = model_b
        self.input_data_b = input_data_b
        self.input_data_a = self.input_data_b if input_data_a is None else input_data_a

        self.id_to_module_a = {id(m): m for m in self.model_a.modules()}
        self.id_to_module_b = {id(m): m for m in self.model_b.modules()}
        self._err_msg = "A and B must have the same model architecture."

        self.model_graph = self.initialize_permutations()

    def initialize_permutations(self) -> PathSequence:
        """Initialize the permutations of the model.
        """

        nextnodes_a: list[NODE_TYPES] = list(
            draw_graph(self.model_a, input_data=self.input_data_a, depth=1e12)
            .root_container
        )
        nextnodes_b: list[NODE_TYPES] = list(
            draw_graph(self.model_b, input_data=self.input_data_b, depth=1e12)
            .root_container
        )

        paths: list[LinearPath | ParallelPaths] = []

        while len(nextnodes_a) > 0 or len(nextnodes_b) > 0:
            assert len(nextnodes_a) == len(nextnodes_b), self._err_msg

            path: LinearPath | ParallelPaths
            if len(nextnodes_b) == 1:
                path, nextnodes_a, nextnodes_b = self.initialize_linear_path(
                    nextnodes_a, nextnodes_b
                )
            else:
                path, nextnodes_a, nextnodes_b = self.initialize_parallel_paths(
                    nextnodes_a, nextnodes_b
                )

            if bool(path):
                # This might be False (path empty) if we, for example,
                #   start with a nn.MultiheadAttention module.
                # Then, the three input tensors would yield a ParallelPaths,
                #   but it would be empty because those tensors combine into
                #   a single Module, the nn.MultiheadAttention.
                # In this case, we aren't interested in the empty ParallelPaths,
                #   only in the subsequent LinearPath (and what follows it).
                paths.append(path)

        return PathSequence(*paths)

    def initialize_linear_path(
            self, nextnodes_a: list[NODE_TYPES], nextnodes_b: list[NODE_TYPES]
    ) -> tuple[LinearPath, list[NODE_TYPES], list[NODE_TYPES]]:

        modules: list[MODULE_TYPES] = []
        while 0 < len(nextnodes_a) < 2:
            assert len(nextnodes_a) == len(nextnodes_b) == 1, self._err_msg

            node_a, node_b = nextnodes_a.pop(), nextnodes_b.pop()
            assert type(node_a) == type(node_b), self._err_msg

            if isinstance(node_a, ModuleNode) and isinstance(node_b, ModuleNode):
                mod_a = self.id_to_module_a[node_a.compute_unit_id]
                mod_b = self.id_to_module_b[node_b.compute_unit_id]
                module = initialize_module(mod_a, mod_b, node_b)
                if module is not None:
                    modules.append(module)

            nextnodes_a = list(node_a.children)  # type: ignore[arg-type]
            nextnodes_b = list(node_b.children)  # type: ignore[arg-type]

            # If there are multiple parents in the next nodes,
            #   then we have reached a fork in the graph.
            # This is likely because initialize_linear_path was called by
            #   initialize_parallel_paths, and we reached the end of the parallel paths.
            if nextnodes_a and len(nextnodes_a[0].parents) > 1:
                assert nextnodes_b and len(nextnodes_b[0].parents) > 1, self._err_msg
                break

        return LinearPath(*modules), nextnodes_a, nextnodes_b

    def initialize_parallel_paths(
            self, nextnodes_a: list[NODE_TYPES], nextnodes_b: list[NODE_TYPES]
    ) -> tuple[ParallelPaths, list[NODE_TYPES], list[NODE_TYPES]]:
        # In some models, there are nested parallel paths.
        # For example, in torchvision.models.efficientnet_b1,
        #   there are parallel paths which consist of a linear path
        #   next to a sequence of a linear path followed by a parallel path,
        #   which is then followed by another linear path.
        # To handle this, we first find the common final nodes of all parallel paths.
        # We can do this with only model_b, because the similarity of the
        #   model architectures is tested later when constructing the paths.
        # This is what we do second: construct the paths.

        # 1. Find common final nodes
        common_finalnodes = self._get_finalnodes(nextnodes_a, nextnodes_b)

        # 2. Construct paths
        paths: list[LinearPath | PathSequence] = []
        finalnodes_a: list[NODE_TYPES] = []
        finalnodes_b: list[NODE_TYPES] = []

        for node_a, node_b in zip(nextnodes_a, nextnodes_b):
            if [node_b] == common_finalnodes:
                # If the node is a common final node,
                #   then it is already handled by the common final nodes.
                paths.append(LinearPath())
                continue
            path, finalnodes_a, finalnodes_b = self._construct_subpath(
                node_a, node_b, common_finalnodes
            )
            assert len(nextnodes_a) == len(nextnodes_b), self._err_msg

            paths.append(path)

        return ParallelPaths(*paths), finalnodes_a, finalnodes_b

    def _get_finalnodes(
            self, nextnodes_a: list[NODE_TYPES], nextnodes_b: list[NODE_TYPES]
    ) -> list[NODE_TYPES]:
        finalnodes_per_path: list[list[list[NODE_TYPES]]] = [
            [[n]] for n in nextnodes_b
        ]

        while True:
            nextnextnodes_a: list[NODE_TYPES | None] = []
            nextnextnodes_b: list[NODE_TYPES | None] = []
            for i, (node_a, node_b) in enumerate(zip(nextnodes_a, nextnodes_b)):
                if node_b is None:
                    assert node_a is None, self._err_msg
                    nextnextnodes_a.append(None)
                    nextnextnodes_b.append(None)
                    continue

                _, finalnodes_a, finalnodes_b = self.initialize_linear_path(
                    [node_a], [node_b]
                )
                finalnodes_per_path[i].append(finalnodes_b)

                # If we have multiple next nodes, we can simply look at the first,
                #   because all nodes should end in one finalnodes list.
                # The specific structure of the subpaths
                #   will be handeled in _construct_subpath.
                nextnextnodes_a.append(finalnodes_a[0] if finalnodes_a else None)
                nextnextnodes_b.append(finalnodes_b[0] if finalnodes_b else None)

            if all(node is None for node in nextnextnodes_b):
                return []

            for finalnodes_b in finalnodes_per_path[0]:
                if all(
                        finalnodes_b in finalnodes_per_path[i]
                        for i in range(1, len(finalnodes_per_path))
                ):
                    return finalnodes_b

            nextnodes_a = nextnextnodes_a  # type: ignore[assignment]
            nextnodes_b = nextnextnodes_b  # type: ignore[assignment]

    def _construct_subpath(
            self,
            node_a: NODE_TYPES,
            node_b: NODE_TYPES,
            common_finalnodes: list[NODE_TYPES]
    ) -> tuple[LinearPath | PathSequence, list[NODE_TYPES], list[NODE_TYPES]]:
        path0, finalnodes_a, finalnodes_b = self.initialize_linear_path(
            [node_a], [node_b]
        )
        paths: list[LinearPath | ParallelPaths] = [path0]

        while finalnodes_b != common_finalnodes:
            assert len(finalnodes_a) == len(finalnodes_b), self._err_msg

            path: LinearPath | ParallelPaths

            if len(finalnodes_a) == 0:
                break
            if len(finalnodes_a) == 1:
                # If there is only one final node, we can simply construct the path
                #   from the next nodes.
                path, finalnodes_a, finalnodes_b = self.initialize_linear_path(
                    finalnodes_a, finalnodes_b
                )
            else:
                # If there are multiple final nodes,
                #   we have to construct a ParallelPaths.
                path, finalnodes_a, finalnodes_b = self.initialize_parallel_paths(
                    finalnodes_a, finalnodes_b
                )
            paths.append(path)

        if len(paths) == 0:
            raise ValueError("No paths found")

        if len(paths) == 1:
            ret_path = paths[0]
            assert isinstance(ret_path, LinearPath), \
                "Can't have PathSequence with one path only"
            return ret_path, finalnodes_a, finalnodes_b

        return PathSequence(*paths), finalnodes_a, finalnodes_b
